# test_refs.py -- tests for refs.py
# Copyright (C) 2013 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Tests for dulwich.refs."""

import errno
import os
import sys
import tempfile
from io import BytesIO
from typing import ClassVar

from dulwich import errors
from dulwich.file import GitFile
from dulwich.objects import ZERO_SHA
from dulwich.protocol import split_peeled_refs, strip_peeled_refs
from dulwich.refs import (
    DictRefsContainer,
    NamespacedRefsContainer,
    SymrefLoop,
    _split_ref_line,
    check_ref_format,
    is_per_worktree_ref,
    parse_remote_ref,
    parse_symref_value,
    read_packed_refs,
    read_packed_refs_with_peeled,
    shorten_ref_name,
    write_packed_refs,
)
from dulwich.repo import Repo
from dulwich.tests.utils import open_repo, tear_down_repo
from dulwich.worktree import add_worktree

from . import SkipTest, TestCase


class CheckRefFormatTests(TestCase):
    """Tests for the check_ref_format function.

    These are the same tests as in the git test suite.
    """

    def test_valid(self) -> None:
        self.assertTrue(check_ref_format(b"heads/foo"))
        self.assertTrue(check_ref_format(b"foo/bar/baz"))
        self.assertTrue(check_ref_format(b"foo./bar"))
        self.assertTrue(check_ref_format(b"heads/foo@bar"))
        self.assertTrue(check_ref_format(b"heads/fix.lock.error"))
        self.assertTrue(check_ref_format(b"heads/foo.lock.bar"))

    def test_invalid(self) -> None:
        self.assertFalse(check_ref_format(b"foo"))
        self.assertFalse(check_ref_format(b"heads/foo/"))
        self.assertFalse(check_ref_format(b"./foo"))
        self.assertFalse(check_ref_format(b".refs/foo"))
        self.assertFalse(check_ref_format(b"heads/foo..bar"))
        self.assertFalse(check_ref_format(b"heads/foo?bar"))
        self.assertFalse(check_ref_format(b"heads/foo.lock"))
        self.assertFalse(check_ref_format(b"heads/v@{ation"))
        self.assertFalse(check_ref_format(b"heads/foo\bar"))
        self.assertFalse(check_ref_format(b"refs//a"))
        self.assertFalse(check_ref_format(b"refs/heads//a"))
        self.assertFalse(check_ref_format(b"refs///heads/foo"))
        self.assertFalse(check_ref_format(b"/refs/heads/foo"))
        self.assertFalse(check_ref_format(b"refs/a.lock/a"))
        self.assertFalse(check_ref_format(b"refs/heads/a.lock/a"))
        self.assertFalse(check_ref_format(b"@"))


ONES = b"1" * 40
TWOS = b"2" * 40
THREES = b"3" * 40
FOURS = b"4" * 40


class PackedRefsFileTests(TestCase):
    def test_split_ref_line_errors(self) -> None:
        self.assertRaises(errors.PackedRefsException, _split_ref_line, b"singlefield")
        self.assertRaises(errors.PackedRefsException, _split_ref_line, b"badsha name")
        self.assertRaises(
            errors.PackedRefsException,
            _split_ref_line,
            ONES + b" bad/../refname",
        )

    def test_read_without_peeled(self) -> None:
        f = BytesIO(b"\n".join([b"# comment", ONES + b" ref/1", TWOS + b" ref/2"]))
        self.assertEqual(
            [(ONES, b"ref/1"), (TWOS, b"ref/2")], list(read_packed_refs(f))
        )

    def test_read_without_peeled_errors(self) -> None:
        f = BytesIO(b"\n".join([ONES + b" ref/1", b"^" + TWOS]))
        self.assertRaises(errors.PackedRefsException, list, read_packed_refs(f))

    def test_read_with_peeled(self) -> None:
        f = BytesIO(
            b"\n".join(
                [
                    ONES + b" ref/1",
                    TWOS + b" ref/2",
                    b"^" + THREES,
                    FOURS + b" ref/4",
                ]
            )
        )
        self.assertEqual(
            [
                (ONES, b"ref/1", None),
                (TWOS, b"ref/2", THREES),
                (FOURS, b"ref/4", None),
            ],
            list(read_packed_refs_with_peeled(f)),
        )

    def test_read_with_peeled_errors(self) -> None:
        f = BytesIO(b"\n".join([b"^" + TWOS, ONES + b" ref/1"]))
        self.assertRaises(errors.PackedRefsException, list, read_packed_refs(f))

        f = BytesIO(b"\n".join([ONES + b" ref/1", b"^" + TWOS, b"^" + THREES]))
        self.assertRaises(errors.PackedRefsException, list, read_packed_refs(f))

    def test_write_with_peeled(self) -> None:
        f = BytesIO()
        write_packed_refs(f, {b"ref/1": ONES, b"ref/2": TWOS}, {b"ref/1": THREES})
        self.assertEqual(
            b"\n".join(
                [
                    b"# pack-refs with: peeled",
                    ONES + b" ref/1",
                    b"^" + THREES,
                    TWOS + b" ref/2",
                ]
            )
            + b"\n",
            f.getvalue(),
        )

    def test_write_without_peeled(self) -> None:
        f = BytesIO()
        write_packed_refs(f, {b"ref/1": ONES, b"ref/2": TWOS})
        self.assertEqual(
            b"\n".join([ONES + b" ref/1", TWOS + b" ref/2"]) + b"\n",
            f.getvalue(),
        )


# Dict of refs that we expect all RefsContainerTests subclasses to define.
_TEST_REFS = {
    b"HEAD": b"42d06bd4b77fed026b154d16493e5deab78f02ec",
    b"refs/heads/40-char-ref-aaaaaaaaaaaaaaaaaa": b"42d06bd4b77fed026b154d16493e5deab78f02ec",
    b"refs/heads/master": b"42d06bd4b77fed026b154d16493e5deab78f02ec",
    b"refs/heads/packed": b"42d06bd4b77fed026b154d16493e5deab78f02ec",
    b"refs/tags/refs-0.1": b"df6800012397fb85c56e7418dd4eb9405dee075c",
    b"refs/tags/refs-0.2": b"3ec9c43c84ff242e3ef4a9fc5bc111fd780a76a8",
    b"refs/heads/loop": b"ref: refs/heads/loop",
}


class RefsContainerTests:
    def test_keys(self) -> None:
        actual_keys = set(self._refs.keys())
        self.assertEqual(set(self._refs.allkeys()), actual_keys)
        self.assertEqual(set(_TEST_REFS.keys()), actual_keys)

        actual_keys = self._refs.keys(b"refs/heads")
        actual_keys.discard(b"loop")
        self.assertEqual(
            [b"40-char-ref-aaaaaaaaaaaaaaaaaa", b"master", b"packed"],
            sorted(actual_keys),
        )
        self.assertEqual(
            [b"refs-0.1", b"refs-0.2"], sorted(self._refs.keys(b"refs/tags"))
        )

    def test_iter(self) -> None:
        actual_keys = set(self._refs.keys())
        self.assertEqual(set(self._refs), actual_keys)
        self.assertEqual(set(_TEST_REFS.keys()), actual_keys)

    def test_as_dict(self) -> None:
        # refs/heads/loop does not show up even if it exists
        expected_refs = dict(_TEST_REFS)
        del expected_refs[b"refs/heads/loop"]
        self.assertEqual(expected_refs, self._refs.as_dict())

    def test_get_symrefs(self) -> None:
        self._refs.set_symbolic_ref(b"refs/heads/src", b"refs/heads/dst")
        symrefs = self._refs.get_symrefs()
        if b"HEAD" in symrefs:
            symrefs.pop(b"HEAD")
        self.assertEqual(
            {
                b"refs/heads/src": b"refs/heads/dst",
                b"refs/heads/loop": b"refs/heads/loop",
            },
            symrefs,
        )

    def test_setitem(self) -> None:
        self._refs[b"refs/some/ref"] = b"42d06bd4b77fed026b154d16493e5deab78f02ec"
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs[b"refs/some/ref"],
        )

        # should accept symref
        self._refs[b"refs/heads/symbolic"] = b"ref: refs/heads/master"
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs[b"refs/heads/symbolic"],
        )

        # should not accept bad ref names
        self.assertRaises(
            errors.RefFormatError,
            self._refs.__setitem__,
            b"notrefs/foo",
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
        )

        # should not accept short sha
        self.assertRaises(
            ValueError,
            self._refs.__setitem__,
            b"refs/some/ref",
            b"42d06bd",
        )

    def test_set_if_equals(self) -> None:
        nines = b"9" * 40
        self.assertFalse(self._refs.set_if_equals(b"HEAD", b"c0ffee", nines))
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec", self._refs[b"HEAD"]
        )

        self.assertTrue(
            self._refs.set_if_equals(
                b"HEAD", b"42d06bd4b77fed026b154d16493e5deab78f02ec", nines
            )
        )
        self.assertEqual(nines, self._refs[b"HEAD"])

        # Setting the ref again is a no-op, but will return True.
        self.assertTrue(self._refs.set_if_equals(b"HEAD", nines, nines))
        self.assertEqual(nines, self._refs[b"HEAD"])

        self.assertTrue(self._refs.set_if_equals(b"refs/heads/master", None, nines))
        self.assertEqual(nines, self._refs[b"refs/heads/master"])

        self.assertTrue(
            self._refs.set_if_equals(b"refs/heads/nonexistent", ZERO_SHA, nines)
        )
        self.assertEqual(nines, self._refs[b"refs/heads/nonexistent"])

    def test_add_if_new(self) -> None:
        nines = b"9" * 40
        self.assertFalse(self._refs.add_if_new(b"refs/heads/master", nines))
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs[b"refs/heads/master"],
        )

        self.assertTrue(self._refs.add_if_new(b"refs/some/ref", nines))
        self.assertEqual(nines, self._refs[b"refs/some/ref"])

    def test_set_symbolic_ref(self) -> None:
        self._refs.set_symbolic_ref(b"refs/heads/symbolic", b"refs/heads/master")
        self.assertEqual(
            b"ref: refs/heads/master",
            self._refs.read_loose_ref(b"refs/heads/symbolic"),
        )
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs[b"refs/heads/symbolic"],
        )

    def test_set_symbolic_ref_overwrite(self) -> None:
        nines = b"9" * 40
        self.assertNotIn(b"refs/heads/symbolic", self._refs)
        self._refs[b"refs/heads/symbolic"] = nines
        self.assertEqual(nines, self._refs.read_loose_ref(b"refs/heads/symbolic"))
        self._refs.set_symbolic_ref(b"refs/heads/symbolic", b"refs/heads/master")
        self.assertEqual(
            b"ref: refs/heads/master",
            self._refs.read_loose_ref(b"refs/heads/symbolic"),
        )
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs[b"refs/heads/symbolic"],
        )

    def test_check_refname(self) -> None:
        self._refs._check_refname(b"HEAD")
        self._refs._check_refname(b"refs/stash")
        self._refs._check_refname(b"refs/heads/foo")

        self.assertRaises(errors.RefFormatError, self._refs._check_refname, b"refs")
        self.assertRaises(
            errors.RefFormatError, self._refs._check_refname, b"notrefs/foo"
        )

    def test_check_refname_empty_component_warns(self) -> None:
        # Names whose only defect is empty path components (e.g. introduced
        # by older Poetry releases) warn rather than raise, for now.
        for name in (b"refs/tags//v1.0", b"refs//heads/foo", b"refs/heads//foo"):
            with self.assertWarns(DeprecationWarning):
                self._refs._check_refname(name)

    def test_check_refname_other_defects_still_raise(self) -> None:
        # Names with defects other than empty components still raise.
        for name in (b"refs/heads//foo.lock", b"refs/heads//foo..bar"):
            self.assertRaises(errors.RefFormatError, self._refs._check_refname, name)

    def test_contains(self) -> None:
        self.assertIn(b"refs/heads/master", self._refs)
        self.assertNotIn(b"refs/heads/bar", self._refs)

    def test_delitem(self) -> None:
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs[b"refs/heads/master"],
        )
        del self._refs[b"refs/heads/master"]
        self.assertRaises(KeyError, lambda: self._refs[b"refs/heads/master"])

    def test_remove_if_equals(self) -> None:
        self.assertFalse(self._refs.remove_if_equals(b"HEAD", b"c0ffee"))
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec", self._refs[b"HEAD"]
        )
        self.assertTrue(
            self._refs.remove_if_equals(
                b"refs/tags/refs-0.2",
                b"3ec9c43c84ff242e3ef4a9fc5bc111fd780a76a8",
            )
        )
        self.assertTrue(self._refs.remove_if_equals(b"refs/tags/refs-0.2", ZERO_SHA))
        self.assertNotIn(b"refs/tags/refs-0.2", self._refs)

    def test_import_refs_name(self) -> None:
        self._refs[b"refs/remotes/origin/other"] = (
            b"48d01bd4b77fed026b154d16493e5deab78f02ec"
        )
        self._refs.import_refs(
            b"refs/remotes/origin",
            {b"master": b"42d06bd4b77fed026b154d16493e5deab78f02ec"},
        )
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs[b"refs/remotes/origin/master"],
        )
        self.assertEqual(
            b"48d01bd4b77fed026b154d16493e5deab78f02ec",
            self._refs[b"refs/remotes/origin/other"],
        )

    def test_import_refs_base_trailing_slash(self) -> None:
        # A trailing slash on the base must not produce a malformed ref
        # with an empty path component (e.g. b'refs/tags//v1.0').
        self._refs.import_refs(
            b"refs/tags/",
            {b"v1.0": b"42d06bd4b77fed026b154d16493e5deab78f02ec"},
        )
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs[b"refs/tags/v1.0"],
        )
        self.assertNotIn(b"refs/tags//v1.0", self._refs.allkeys())

    def test_import_refs_name_prune(self) -> None:
        self._refs[b"refs/remotes/origin/other"] = (
            b"48d01bd4b77fed026b154d16493e5deab78f02ec"
        )
        self._refs.import_refs(
            b"refs/remotes/origin",
            {b"master": b"42d06bd4b77fed026b154d16493e5deab78f02ec"},
            prune=True,
        )
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs[b"refs/remotes/origin/master"],
        )
        self.assertNotIn(b"refs/remotes/origin/other", self._refs)


class DictRefsContainerTests(RefsContainerTests, TestCase):
    def setUp(self) -> None:
        TestCase.setUp(self)
        self._refs = DictRefsContainer(dict(_TEST_REFS))

    def test_invalid_refname(self) -> None:
        # FIXME: Move this test into RefsContainerTests, but requires
        # some way of injecting invalid refs.
        self._refs._refs[b"refs/stash"] = b"00" * 20
        expected_refs = dict(_TEST_REFS)
        del expected_refs[b"refs/heads/loop"]
        expected_refs[b"refs/stash"] = b"00" * 20
        self.assertEqual(expected_refs, self._refs.as_dict())

    def test_set_if_equals_with_symbolic_ref(self) -> None:
        # Test that set_if_equals only updates the requested ref,
        # not all refs in a symbolic reference chain

        # The bug in the original implementation was that when follow()
        # was called on a ref, it would return all refs in the chain,
        # and set_if_equals would update ALL of them instead of just the
        # requested ref.

        # Set up refs
        master_sha = b"1" * 40
        feature_sha = b"2" * 40
        new_sha = b"3" * 40

        self._refs[b"refs/heads/master"] = master_sha
        self._refs[b"refs/heads/feature"] = feature_sha
        # Create a second symbolic ref pointing to feature
        self._refs.set_symbolic_ref(b"refs/heads/other", b"refs/heads/feature")

        # Update refs/heads/other through set_if_equals
        # With the bug, this would update BOTH refs/heads/other AND refs/heads/feature
        # Without the bug, only refs/heads/other should be updated
        # Note: old_ref needs to be the actual stored value (the symref)
        self.assertTrue(
            self._refs.set_if_equals(
                b"refs/heads/other", b"ref: refs/heads/feature", new_sha
            )
        )

        # refs/heads/other should now directly point to new_sha
        self.assertEqual(self._refs.read_ref(b"refs/heads/other"), new_sha)

        # refs/heads/feature should remain unchanged
        # With the bug, refs/heads/feature would also be incorrectly updated to new_sha
        self.assertEqual(self._refs[b"refs/heads/feature"], feature_sha)
        self.assertEqual(self._refs[b"refs/heads/master"], master_sha)


class DiskRefsContainerTests(RefsContainerTests, TestCase):
    def setUp(self) -> None:
        TestCase.setUp(self)
        self._repo = open_repo("refs.git")
        self.addCleanup(tear_down_repo, self._repo)
        self._refs = self._repo.refs

    def test_get_packed_refs(self) -> None:
        self.assertEqual(
            {
                b"refs/heads/packed": b"42d06bd4b77fed026b154d16493e5deab78f02ec",
                b"refs/tags/refs-0.1": b"df6800012397fb85c56e7418dd4eb9405dee075c",
            },
            self._refs.get_packed_refs(),
        )

    def test_get_peeled_not_packed(self) -> None:
        # not packed
        self.assertEqual(None, self._refs.get_peeled(b"refs/tags/refs-0.2"))
        self.assertEqual(
            b"3ec9c43c84ff242e3ef4a9fc5bc111fd780a76a8",
            self._refs[b"refs/tags/refs-0.2"],
        )

        # packed, known not peelable
        self.assertEqual(
            self._refs[b"refs/heads/packed"],
            self._refs.get_peeled(b"refs/heads/packed"),
        )

        # packed, peeled
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs.get_peeled(b"refs/tags/refs-0.1"),
        )

    def test_setitem(self) -> None:
        RefsContainerTests.test_setitem(self)
        path = os.path.join(self._refs.path, b"refs", b"some", b"ref")
        with open(path, "rb") as f:
            self.assertEqual(b"42d06bd4b77fed026b154d16493e5deab78f02ec", f.read()[:40])

        self.assertRaises(
            OSError,
            self._refs.__setitem__,
            b"refs/some/ref/sub",
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
        )

    def test_read_loose_ref_escape(self) -> None:
        # A ref name must not be able to address a file outside the ref store.
        outside = os.path.join(os.path.dirname(self._repo.path), "secret")
        with open(outside, "wb") as f:
            f.write(b"42d06bd4b77fed026b154d16493e5deab78f02ec\n")
        self.assertEqual(None, self._refs.read_loose_ref(b"../secret"))
        self.assertRaises(KeyError, lambda: self._refs[b"../secret"])
        self.assertEqual(None, self._refs.read_loose_ref(b"/etc/passwd"))

    def test_delete_refs_container(self) -> None:
        # We shouldn't delete the refs directory
        self._refs[b"refs/heads/blah"] = b"42d06bd4b77fed026b154d16493e5deab78f02ec"
        for ref in self._refs.allkeys():
            del self._refs[ref]
        self.assertTrue(os.path.exists(os.path.join(self._refs.path, b"refs")))

    def test_setitem_packed(self) -> None:
        with open(os.path.join(self._refs.path, b"packed-refs"), "w") as f:
            f.write("# pack-refs with: peeled fully-peeled sorted \n")
            f.write("42d06bd4b77fed026b154d16493e5deab78f02ec refs/heads/packed\n")

        # It's allowed to set a new ref on a packed ref, the new ref will be
        # placed outside on refs/
        self._refs[b"refs/heads/packed"] = b"3ec9c43c84ff242e3ef4a9fc5bc111fd780a76a8"
        packed_ref_path = os.path.join(self._refs.path, b"refs", b"heads", b"packed")
        with open(packed_ref_path, "rb") as f:
            self.assertEqual(b"3ec9c43c84ff242e3ef4a9fc5bc111fd780a76a8", f.read()[:40])

        self.assertRaises(
            OSError,
            self._refs.__setitem__,
            b"refs/heads/packed/sub",
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
        )

        # this shouldn't overwrite the packed refs
        self.assertEqual(
            {b"refs/heads/packed": b"42d06bd4b77fed026b154d16493e5deab78f02ec"},
            self._refs.get_packed_refs(),
        )

    def test_add_packed_refs(self) -> None:
        # first, create a non-packed ref
        self._refs[b"refs/heads/packed"] = b"3ec9c43c84ff242e3ef4a9fc5bc111fd780a76a8"

        packed_ref_path = os.path.join(self._refs.path, b"refs", b"heads", b"packed")
        self.assertTrue(os.path.exists(packed_ref_path))

        # now overwrite that with a packed ref
        packed_refs_file_path = os.path.join(self._refs.path, b"packed-refs")
        self._refs.add_packed_refs(
            {
                b"refs/heads/packed": b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            }
        )

        # that should kill the file
        self.assertFalse(os.path.exists(packed_ref_path))

        # now delete the packed ref
        self._refs.add_packed_refs(
            {
                b"refs/heads/packed": None,
            }
        )

        # and it's gone!
        self.assertFalse(os.path.exists(packed_ref_path))

        self.assertRaises(
            KeyError,
            self._refs.__getitem__,
            b"refs/heads/packed",
        )

        # just in case, make sure we can't pack HEAD
        self.assertRaises(
            ValueError,
            self._refs.add_packed_refs,
            {b"HEAD": "02ac81614bcdbd585a37b4b0edf8cb8a"},
        )

        # delete all packed refs
        self._refs.add_packed_refs({ref: None for ref in self._refs.get_packed_refs()})

        self.assertEqual({}, self._refs.get_packed_refs())

        # remove the packed ref file, and check that adding nothing doesn't affect that
        os.remove(packed_refs_file_path)

        # adding nothing doesn't make it reappear
        self._refs.add_packed_refs({})

        self.assertFalse(os.path.exists(packed_refs_file_path))

    def test_setitem_symbolic(self) -> None:
        ones = b"1" * 40
        self._refs[b"HEAD"] = ones
        self.assertEqual(ones, self._refs[b"HEAD"])

        # ensure HEAD was not modified
        f = open(os.path.join(self._refs.path, b"HEAD"), "rb")
        v = next(iter(f)).rstrip(b"\n\r")
        f.close()
        self.assertEqual(b"ref: refs/heads/master", v)

        # ensure the symbolic link was written through
        f = open(os.path.join(self._refs.path, b"refs", b"heads", b"master"), "rb")
        self.assertEqual(ones, f.read()[:40])
        f.close()

    def test_set_if_equals(self) -> None:
        RefsContainerTests.test_set_if_equals(self)

        # ensure symref was followed
        self.assertEqual(b"9" * 40, self._refs[b"refs/heads/master"])

        # ensure lockfile was deleted
        self.assertFalse(
            os.path.exists(
                os.path.join(self._refs.path, b"refs", b"heads", b"master.lock")
            )
        )
        self.assertFalse(os.path.exists(os.path.join(self._refs.path, b"HEAD.lock")))

    def test_add_if_new_packed(self) -> None:
        # don't overwrite packed ref
        self.assertFalse(self._refs.add_if_new(b"refs/tags/refs-0.1", b"9" * 40))
        self.assertEqual(
            b"df6800012397fb85c56e7418dd4eb9405dee075c",
            self._refs[b"refs/tags/refs-0.1"],
        )

    def test_add_if_new_symbolic(self) -> None:
        # Use an empty repo instead of the default.
        repo_dir = os.path.join(tempfile.mkdtemp(), "test")
        os.makedirs(repo_dir)
        repo = Repo.init(repo_dir)
        self.addCleanup(tear_down_repo, repo)
        refs = repo.refs

        nines = b"9" * 40
        self.assertEqual(b"ref: refs/heads/master", refs.read_ref(b"HEAD"))
        self.assertNotIn(b"refs/heads/master", refs)
        self.assertTrue(refs.add_if_new(b"HEAD", nines))
        self.assertEqual(b"ref: refs/heads/master", refs.read_ref(b"HEAD"))
        self.assertEqual(nines, refs[b"HEAD"])
        self.assertEqual(nines, refs[b"refs/heads/master"])
        self.assertFalse(refs.add_if_new(b"HEAD", b"1" * 40))
        self.assertEqual(nines, refs[b"HEAD"])
        self.assertEqual(nines, refs[b"refs/heads/master"])

    def test_follow(self) -> None:
        self.assertEqual(
            (
                [b"HEAD", b"refs/heads/master"],
                b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            ),
            self._refs.follow(b"HEAD"),
        )
        self.assertEqual(
            (
                [b"refs/heads/master"],
                b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            ),
            self._refs.follow(b"refs/heads/master"),
        )
        self.assertRaises(SymrefLoop, self._refs.follow, b"refs/heads/loop")

    def test_set_overwrite_loop(self) -> None:
        self.assertRaises(SymrefLoop, self._refs.follow, b"refs/heads/loop")
        self._refs[b"refs/heads/loop"] = b"42d06bd4b77fed026b154d16493e5deab78f02ec"
        self.assertEqual(
            ([b"refs/heads/loop"], b"42d06bd4b77fed026b154d16493e5deab78f02ec"),
            self._refs.follow(b"refs/heads/loop"),
        )

    def test_delitem(self) -> None:
        RefsContainerTests.test_delitem(self)
        ref_file = os.path.join(self._refs.path, b"refs", b"heads", b"master")
        self.assertFalse(os.path.exists(ref_file))
        self.assertNotIn(b"refs/heads/master", self._refs.get_packed_refs())

    def test_delitem_symbolic(self) -> None:
        self.assertEqual(b"ref: refs/heads/master", self._refs.read_loose_ref(b"HEAD"))
        del self._refs[b"HEAD"]
        self.assertRaises(KeyError, lambda: self._refs[b"HEAD"])
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs[b"refs/heads/master"],
        )
        self.assertFalse(os.path.exists(os.path.join(self._refs.path, b"HEAD")))

    def test_remove_if_equals_symref(self) -> None:
        # HEAD is a symref, so shouldn't equal its dereferenced value
        self.assertFalse(
            self._refs.remove_if_equals(
                b"HEAD", b"42d06bd4b77fed026b154d16493e5deab78f02ec"
            )
        )
        self.assertTrue(
            self._refs.remove_if_equals(
                b"refs/heads/master",
                b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            )
        )
        self.assertRaises(KeyError, lambda: self._refs[b"refs/heads/master"])

        # HEAD is now a broken symref
        self.assertRaises(KeyError, lambda: self._refs[b"HEAD"])
        self.assertEqual(b"ref: refs/heads/master", self._refs.read_loose_ref(b"HEAD"))

        self.assertFalse(
            os.path.exists(
                os.path.join(self._refs.path, b"refs", b"heads", b"master.lock")
            )
        )
        self.assertFalse(os.path.exists(os.path.join(self._refs.path, b"HEAD.lock")))

    def test_remove_packed_without_peeled(self) -> None:
        refs_file = os.path.join(self._repo.path, "packed-refs")
        f = GitFile(refs_file)
        refs_data = f.read()
        f.close()
        f = GitFile(refs_file, "wb")
        f.write(
            b"\n".join(
                line
                for line in refs_data.split(b"\n")
                if not line or line[0] not in b"#^"
            )
        )
        f.close()
        self._repo = Repo(self._repo.path)
        refs = self._repo.refs
        self.assertTrue(
            refs.remove_if_equals(
                b"refs/heads/packed",
                b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            )
        )

    def test_remove_if_equals_packed(self) -> None:
        # test removing ref that is only packed
        self.assertEqual(
            b"df6800012397fb85c56e7418dd4eb9405dee075c",
            self._refs[b"refs/tags/refs-0.1"],
        )
        self.assertTrue(
            self._refs.remove_if_equals(
                b"refs/tags/refs-0.1",
                b"df6800012397fb85c56e7418dd4eb9405dee075c",
            )
        )
        self.assertRaises(KeyError, lambda: self._refs[b"refs/tags/refs-0.1"])

    def test_remove_parent(self) -> None:
        self._refs[b"refs/heads/foo/bar"] = b"df6800012397fb85c56e7418dd4eb9405dee075c"
        del self._refs[b"refs/heads/foo/bar"]
        ref_file = os.path.join(
            self._refs.path,
            b"refs",
            b"heads",
            b"foo",
            b"bar",
        )
        self.assertFalse(os.path.exists(ref_file))
        ref_file = os.path.join(self._refs.path, b"refs", b"heads", b"foo")
        self.assertFalse(os.path.exists(ref_file))
        ref_file = os.path.join(self._refs.path, b"refs", b"heads")
        self.assertTrue(os.path.exists(ref_file))
        self._refs[b"refs/heads/foo"] = b"df6800012397fb85c56e7418dd4eb9405dee075c"

    def test_read_ref(self) -> None:
        self.assertEqual(b"ref: refs/heads/master", self._refs.read_ref(b"HEAD"))
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs.read_ref(b"refs/heads/packed"),
        )
        self.assertEqual(None, self._refs.read_ref(b"nonexistent"))

    def test_read_loose_ref(self) -> None:
        self._refs[b"refs/heads/foo"] = b"df6800012397fb85c56e7418dd4eb9405dee075c"

        self.assertEqual(None, self._refs.read_ref(b"refs/heads/foo/bar"))

    def test_non_ascii(self) -> None:
        try:
            encoded_ref = os.fsencode("refs/tags/schön")
        except UnicodeEncodeError as exc:
            raise SkipTest(
                "filesystem encoding doesn't support special character"
            ) from exc
        p = os.path.join(os.fsencode(self._repo.path), encoded_ref)
        with open(p, "w") as f:
            f.write("00" * 20)

        expected_refs = dict(_TEST_REFS)
        expected_refs[encoded_ref] = b"00" * 20
        del expected_refs[b"refs/heads/loop"]

        self.assertEqual(expected_refs, self._repo.get_refs())

    def test_cyrillic(self) -> None:
        if sys.platform in ("darwin", "win32"):
            raise SkipTest("filesystem encoding doesn't support arbitrary bytes")
        # reported in https://github.com/dulwich/dulwich/issues/608
        name = b"\xcd\xee\xe2\xe0\xff\xe2\xe5\xf2\xea\xe01"
        encoded_ref = b"refs/heads/" + name
        try:
            with open(
                os.path.join(os.fsencode(self._repo.path), encoded_ref), "w"
            ) as f:
                f.write("00" * 20)
        except OSError as e:
            if e.errno == errno.EILSEQ:
                # Filesystem rejects non-UTF8 filenames (e.g. ZFS utf8only=on).
                raise SkipTest(
                    f"filesystem rejects non-UTF8 filename {e.filename!r}"
                ) from e
            raise

        expected_refs = set(_TEST_REFS.keys())
        expected_refs.add(encoded_ref)

        self.assertEqual(expected_refs, set(self._repo.refs.allkeys()))
        self.assertEqual(
            {r[len(b"refs/") :] for r in expected_refs if r.startswith(b"refs/")},
            set(self._repo.refs.subkeys(b"refs/")),
        )
        expected_refs.remove(b"refs/heads/loop")
        expected_refs.add(b"HEAD")
        self.assertEqual(expected_refs, set(self._repo.get_refs().keys()))

    def test_write_unchanged_ref_optimization(self):
        # Test that writing unchanged ref avoids fsync but still checks locks
        ref_name = b"refs/heads/unchanged"
        ref_value = b"a" * 40

        # Set initial ref value
        self._refs[ref_name] = ref_value

        # Test 1: Writing same value should succeed without changes
        result = self._refs.set_if_equals(ref_name, ref_value, ref_value)
        self.assertTrue(result)

        # Test 2: Writing same value with wrong old_ref should fail
        wrong_old = b"b" * 40
        result = self._refs.set_if_equals(ref_name, wrong_old, ref_value)
        self.assertFalse(result)

        # Test 3: Writing different value should update normally
        new_value = b"c" * 40
        result = self._refs.set_if_equals(ref_name, ref_value, new_value)
        self.assertTrue(result)
        self.assertEqual(new_value, self._refs[ref_name])

    def test_write_unchanged_ref_with_lock(self):
        # Test that file locking is still detected when ref unchanged

        from dulwich.file import FileLocked

        ref_name = b"refs/heads/locktest"
        ref_value = b"d" * 40

        # Set initial ref value
        self._refs[ref_name] = ref_value

        # Get the actual file path
        ref_file = os.path.join(os.fsencode(self._refs.path), ref_name)
        lock_file = ref_file + b".lock"

        # Create lock file to simulate another process holding lock
        with open(lock_file, "wb") as f:
            f.write(b"locked by another process")

            # Try to write same value - should raise FileLocked
            with self.assertRaises(FileLocked):
                self._refs[ref_name] = ref_value

        # Clean up lock file
        if os.path.exists(lock_file):
            os.unlink(lock_file)

        # Now it should work
        self._refs[ref_name] = ref_value


class IsPerWorktreeRefsTests(TestCase):
    def test(self) -> None:
        cases = [
            (b"HEAD", True),
            (b"refs/bisect/good", True),
            (b"refs/worktree/foo", True),
            (b"refs/rewritten/onto", True),
            (b"refs/stash", False),
            (b"refs/heads/main", False),
            (b"refs/tags/v1.0", False),
            (b"refs/remotes/origin/main", False),
            (b"refs/custom/foo", False),
            (b"refs/replace/aaaaaa", False),
        ]
        for ref, expected in cases:
            with self.subTest(ref=ref, expected=expected):
                self.assertEqual(is_per_worktree_ref(ref), expected)


class DiskRefsContainerWorktreeRefsTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        # Create temporary directories
        temp_dir = tempfile.mkdtemp()
        test_dir = os.path.join(temp_dir, "main")
        os.makedirs(test_dir)

        repo = Repo.init(test_dir, default_branch=b"main")
        main_worktree = repo.get_worktree()
        with open(os.path.join(test_dir, "test.txt"), "wb") as f:
            f.write(b"test content")
        main_worktree.stage(["test.txt"])
        self.first_commit = main_worktree.commit(message=b"Initial commit")

        worktree_dir = os.path.join(temp_dir, "worktree")
        wt_repo = add_worktree(repo, worktree_dir, branch="wt-main")
        linked_worktree = wt_repo.get_worktree()
        with open(os.path.join(test_dir, "test2.txt"), "wb") as f:
            f.write(b"test content")
        linked_worktree.stage(["test2.txt"])
        self.second_commit = linked_worktree.commit(message=b"second commit")

        self.refs = repo.refs
        self.wt_refs = wt_repo.refs

    def test_refpath(self) -> None:
        main_path = self.refs.path
        common = self.wt_refs.path
        wt_path = self.wt_refs.worktree_path

        cases = [
            (self.refs, b"HEAD", main_path),
            (self.refs, b"refs/heads/main", main_path),
            (self.refs, b"refs/heads/wt-main", main_path),
            (self.refs, b"refs/worktree/foo", main_path),
            (self.refs, b"refs/bisect/good", main_path),
            (self.wt_refs, b"HEAD", wt_path),
            (self.wt_refs, b"refs/heads/main", common),
            (self.wt_refs, b"refs/heads/wt-main", common),
            (self.wt_refs, b"refs/worktree/foo", wt_path),
            (self.wt_refs, b"refs/bisect/good", wt_path),
        ]

        for refs, refname, git_dir in cases:
            with self.subTest(refs=refs, refname=refname, git_dir=git_dir):
                refpath = refs.refpath(refname)
                expected_path = os.path.join(
                    git_dir, refname.replace(b"/", os.fsencode(os.sep))
                )
                self.assertEqual(refpath, expected_path)

    def test_shared_ref(self) -> None:
        self.assertEqual(self.refs[b"refs/heads/main"], self.first_commit)
        self.assertEqual(self.refs[b"refs/heads/wt-main"], self.second_commit)
        self.assertEqual(self.wt_refs[b"refs/heads/main"], self.first_commit)
        self.assertEqual(self.wt_refs[b"refs/heads/wt-main"], self.second_commit)

        expected = {b"HEAD", b"refs/heads/main", b"refs/heads/wt-main"}
        self.assertEqual(expected, self.refs.keys())
        self.assertEqual(expected, self.wt_refs.keys())

        self.assertEqual({b"main", b"wt-main"}, set(self.refs.keys(b"refs/heads/")))
        self.assertEqual({b"main", b"wt-main"}, set(self.wt_refs.keys(b"refs/heads/")))

        ref_path = os.path.join(self.refs.path, b"refs", b"heads", b"main")
        self.assertTrue(os.path.exists(ref_path))

        ref_path = os.path.join(self.wt_refs.worktree_path, b"refs", b"heads", b"main")
        self.assertFalse(os.path.exists(ref_path))

    def test_per_worktree_ref(self) -> None:
        path = self.refs.path
        wt_path = self.wt_refs.worktree_path

        self.assertEqual(self.refs[b"HEAD"], self.first_commit)
        self.assertEqual(self.wt_refs[b"HEAD"], self.second_commit)

        self.refs[b"refs/bisect/good"] = self.first_commit
        self.wt_refs[b"refs/bisect/good"] = self.second_commit

        self.refs[b"refs/bisect/start"] = self.first_commit
        self.wt_refs[b"refs/bisect/bad"] = self.second_commit

        self.assertEqual(self.refs[b"refs/bisect/good"], self.first_commit)
        self.assertEqual(self.wt_refs[b"refs/bisect/good"], self.second_commit)

        self.assertTrue(os.path.exists(os.path.join(path, b"refs", b"bisect", b"good")))
        self.assertTrue(
            os.path.exists(os.path.join(wt_path, b"refs", b"bisect", b"good"))
        )

        self.assertEqual(self.refs[b"refs/bisect/start"], self.first_commit)
        with self.assertRaises(KeyError):
            self.wt_refs[b"refs/bisect/start"]
        self.assertTrue(
            os.path.exists(os.path.join(path, b"refs", b"bisect", b"start"))
        )
        self.assertFalse(
            os.path.exists(os.path.join(wt_path, b"refs", b"bisect", b"start"))
        )

        with self.assertRaises(KeyError):
            self.refs[b"refs/bisect/bad"]
        self.assertEqual(self.wt_refs[b"refs/bisect/bad"], self.second_commit)
        self.assertFalse(os.path.exists(os.path.join(path, b"refs", b"bisect", b"bad")))
        self.assertTrue(
            os.path.exists(os.path.join(wt_path, b"refs", b"bisect", b"bad"))
        )

        expected_refs = {
            b"HEAD",
            b"refs/heads/main",
            b"refs/heads/wt-main",
            b"refs/bisect/good",
            b"refs/bisect/start",
        }
        self.assertEqual(self.refs.keys(), expected_refs)
        self.assertEqual({b"good", b"start"}, self.refs.keys(b"refs/bisect/"))

        expected_wt_refs = {
            b"HEAD",
            b"refs/heads/main",
            b"refs/heads/wt-main",
            b"refs/bisect/good",
            b"refs/bisect/bad",
        }
        self.assertEqual(self.wt_refs.keys(), expected_wt_refs)
        self.assertEqual({b"good", b"bad"}, self.wt_refs.keys(b"refs/bisect/"))

    def test_delete_per_worktree_ref(self) -> None:
        self.refs[b"refs/worktree/foo"] = self.first_commit
        self.wt_refs[b"refs/worktree/foo"] = self.second_commit

        del self.wt_refs[b"refs/worktree/foo"]
        with self.assertRaises(KeyError):
            self.wt_refs[b"refs/worktree/foo"]

        del self.refs[b"refs/worktree/foo"]
        with self.assertRaises(KeyError):
            self.refs[b"refs/worktree/foo"]

    def test_delete_shared_ref(self) -> None:
        self.refs[b"refs/heads/branch"] = self.first_commit

        del self.wt_refs[b"refs/heads/branch"]

        with self.assertRaises(KeyError):
            self.wt_refs[b"refs/heads/branch"]
        with self.assertRaises(KeyError):
            self.refs[b"refs/heads/branch"]

    def test_contains_shared_ref(self):
        self.assertIn(b"refs/heads/main", self.refs)
        self.assertIn(b"refs/heads/main", self.wt_refs)
        self.assertIn(b"refs/heads/wt-main", self.refs)
        self.assertIn(b"refs/heads/wt-main", self.wt_refs)

    def test_contains_per_worktree_ref(self):
        self.refs[b"refs/worktree/foo"] = self.first_commit
        self.wt_refs[b"refs/worktree/bar"] = self.second_commit

        self.assertIn(b"refs/worktree/foo", self.refs)
        self.assertNotIn(b"refs/worktree/bar", self.refs)
        self.assertNotIn(b"refs/worktree/foo", self.wt_refs)
        self.assertIn(b"refs/worktree/bar", self.wt_refs)


_TEST_REFS_SERIALIZED = (
    b"42d06bd4b77fed026b154d16493e5deab78f02ec\t"
    b"refs/heads/40-char-ref-aaaaaaaaaaaaaaaaaa\n"
    b"42d06bd4b77fed026b154d16493e5deab78f02ec\trefs/heads/master\n"
    b"42d06bd4b77fed026b154d16493e5deab78f02ec\trefs/heads/packed\n"
    b"df6800012397fb85c56e7418dd4eb9405dee075c\trefs/tags/refs-0.1\n"
    b"3ec9c43c84ff242e3ef4a9fc5bc111fd780a76a8\trefs/tags/refs-0.2\n"
)


class DiskRefsContainerPathlibTests(TestCase):
    def test_pathlib_init(self) -> None:
        from pathlib import Path

        from dulwich.refs import DiskRefsContainer

        # Create a temporary directory
        temp_dir = tempfile.mkdtemp()
        self.addCleanup(os.rmdir, temp_dir)

        # Test with pathlib.Path
        path_obj = Path(temp_dir)
        refs = DiskRefsContainer(path_obj)
        self.assertEqual(refs.path, temp_dir.encode())

        # Test refpath with pathlib initialized container
        ref_path = refs.refpath(b"HEAD")
        self.assertTrue(isinstance(ref_path, bytes))
        self.assertEqual(ref_path, os.path.join(temp_dir.encode(), b"HEAD"))

    def test_pathlib_worktree_path(self) -> None:
        from pathlib import Path

        from dulwich.refs import DiskRefsContainer

        # Create temporary directories
        temp_dir = tempfile.mkdtemp()
        worktree_dir = tempfile.mkdtemp()
        self.addCleanup(os.rmdir, temp_dir)
        self.addCleanup(os.rmdir, worktree_dir)

        # Test with pathlib.Path for both paths
        path_obj = Path(temp_dir)
        worktree_obj = Path(worktree_dir)
        refs = DiskRefsContainer(path_obj, worktree_path=worktree_obj)
        self.assertEqual(refs.path, temp_dir.encode())
        self.assertEqual(refs.worktree_path, worktree_dir.encode())

        # Test refpath returns worktree path for HEAD
        ref_path = refs.refpath(b"HEAD")
        self.assertEqual(ref_path, os.path.join(worktree_dir.encode(), b"HEAD"))


class ParseSymrefValueTests(TestCase):
    def test_valid(self) -> None:
        self.assertEqual(b"refs/heads/foo", parse_symref_value(b"ref: refs/heads/foo"))

    def test_invalid(self) -> None:
        self.assertRaises(ValueError, parse_symref_value, b"foobar")


class ParseRemoteRefTests(TestCase):
    def test_valid(self) -> None:
        # Test simple case
        remote, branch = parse_remote_ref(b"refs/remotes/origin/main")
        self.assertEqual(b"origin", remote)
        self.assertEqual(b"main", branch)

        # Test with branch containing slashes
        remote, branch = parse_remote_ref(b"refs/remotes/upstream/feature/new-ui")
        self.assertEqual(b"upstream", remote)
        self.assertEqual(b"feature/new-ui", branch)

    def test_invalid_not_remote_ref(self) -> None:
        # Not a remote ref
        with self.assertRaises(ValueError) as cm:
            parse_remote_ref(b"refs/heads/main")
        self.assertIn("Not a remote ref", str(cm.exception))

    def test_invalid_format(self) -> None:
        # Missing branch name
        with self.assertRaises(ValueError) as cm:
            parse_remote_ref(b"refs/remotes/origin")
        self.assertIn("Invalid remote ref format", str(cm.exception))

        # Just the prefix
        with self.assertRaises(ValueError) as cm:
            parse_remote_ref(b"refs/remotes/")
        self.assertIn("Invalid remote ref format", str(cm.exception))


class StripPeeledRefsTests(TestCase):
    all_refs: ClassVar[dict[bytes, bytes]] = {
        b"refs/heads/master": b"8843d7f92416211de9ebb963ff4ce28125932878",
        b"refs/heads/testing": b"186a005b134d8639a58b6731c7c1ea821a6eedba",
        b"refs/tags/1.0.0": b"a93db4b0360cc635a2b93675010bac8d101f73f0",
        b"refs/tags/1.0.0^{}": b"a93db4b0360cc635a2b93675010bac8d101f73f0",
        b"refs/tags/2.0.0": b"0749936d0956c661ac8f8d3483774509c165f89e",
        b"refs/tags/2.0.0^{}": b"0749936d0956c661ac8f8d3483774509c165f89e",
    }
    non_peeled_refs: ClassVar[dict[bytes, bytes]] = {
        b"refs/heads/master": b"8843d7f92416211de9ebb963ff4ce28125932878",
        b"refs/heads/testing": b"186a005b134d8639a58b6731c7c1ea821a6eedba",
        b"refs/tags/1.0.0": b"a93db4b0360cc635a2b93675010bac8d101f73f0",
        b"refs/tags/2.0.0": b"0749936d0956c661ac8f8d3483774509c165f89e",
    }

    def test_strip_peeled_refs(self) -> None:
        # Simple check of two dicts
        self.assertEqual(strip_peeled_refs(self.all_refs), self.non_peeled_refs)

    def test_split_peeled_refs(self) -> None:
        (regular, peeled) = split_peeled_refs(self.all_refs)
        self.assertEqual(regular, self.non_peeled_refs)
        self.assertEqual(
            peeled,
            {
                b"refs/tags/2.0.0": b"0749936d0956c661ac8f8d3483774509c165f89e",
                b"refs/tags/1.0.0": b"a93db4b0360cc635a2b93675010bac8d101f73f0",
            },
        )


class ShortenRefNameTests(TestCase):
    """Tests for shorten_ref_name function."""

    def test_branch_ref(self) -> None:
        """Test shortening branch references."""
        self.assertEqual(b"master", shorten_ref_name(b"refs/heads/master"))
        self.assertEqual(b"develop", shorten_ref_name(b"refs/heads/develop"))
        self.assertEqual(
            b"feature/new-ui", shorten_ref_name(b"refs/heads/feature/new-ui")
        )

    def test_remote_ref(self) -> None:
        """Test shortening remote references."""
        self.assertEqual(b"origin/main", shorten_ref_name(b"refs/remotes/origin/main"))
        self.assertEqual(
            b"upstream/master", shorten_ref_name(b"refs/remotes/upstream/master")
        )
        self.assertEqual(
            b"origin/feature/test",
            shorten_ref_name(b"refs/remotes/origin/feature/test"),
        )

    def test_tag_ref(self) -> None:
        """Test shortening tag references."""
        self.assertEqual(b"v1.0", shorten_ref_name(b"refs/tags/v1.0"))
        self.assertEqual(b"release-2.0", shorten_ref_name(b"refs/tags/release-2.0"))

    def test_special_refs(self) -> None:
        """Test that special refs are not shortened."""
        self.assertEqual(b"HEAD", shorten_ref_name(b"HEAD"))
        self.assertEqual(b"FETCH_HEAD", shorten_ref_name(b"FETCH_HEAD"))
        self.assertEqual(b"ORIG_HEAD", shorten_ref_name(b"ORIG_HEAD"))

    def test_other_refs(self) -> None:
        """Test refs that don't match standard prefixes."""
        # Refs that don't match any standard prefix are returned as-is
        self.assertEqual(b"refs/stash", shorten_ref_name(b"refs/stash"))
        self.assertEqual(b"refs/bisect/good", shorten_ref_name(b"refs/bisect/good"))


class RefUtilityFunctionsTests(TestCase):
    """Tests for the new ref utility functions."""

    def test_local_branch_name(self) -> None:
        """Test local_branch_name function."""
        from dulwich.refs import local_branch_name

        # Test adding prefix to branch name
        self.assertEqual(b"refs/heads/master", local_branch_name(b"master"))
        self.assertEqual(b"refs/heads/develop", local_branch_name(b"develop"))
        self.assertEqual(
            b"refs/heads/feature/new-ui", local_branch_name(b"feature/new-ui")
        )

        # Test idempotency - already has prefix
        self.assertEqual(b"refs/heads/master", local_branch_name(b"refs/heads/master"))

    def test_local_branch_name_leading_slash(self) -> None:
        """Test local_branch_name warns about and strips a leading slash."""
        from dulwich.refs import local_branch_name

        with self.assertWarns(DeprecationWarning):
            self.assertEqual(b"refs/heads/master", local_branch_name(b"/master"))

    def test_local_tag_name(self) -> None:
        """Test local_tag_name function."""
        from dulwich.refs import local_tag_name

        # Test adding prefix to tag name
        self.assertEqual(b"refs/tags/v1.0", local_tag_name(b"v1.0"))
        self.assertEqual(b"refs/tags/release-2.0", local_tag_name(b"release-2.0"))

        # Test idempotency - already has prefix
        self.assertEqual(b"refs/tags/v1.0", local_tag_name(b"refs/tags/v1.0"))

    def test_local_tag_name_leading_slash(self) -> None:
        """Test local_tag_name warns about and strips a leading slash."""
        from dulwich.refs import local_tag_name

        with self.assertWarns(DeprecationWarning):
            self.assertEqual(b"refs/tags/v1.0", local_tag_name(b"/v1.0"))

    def test_local_replace_name_leading_slash(self) -> None:
        """Test local_replace_name warns about and strips a leading slash."""
        from dulwich.refs import local_replace_name

        with self.assertWarns(DeprecationWarning):
            self.assertEqual(b"refs/replace/abc123", local_replace_name(b"/abc123"))

    def test_extract_branch_name(self) -> None:
        """Test extract_branch_name function."""
        from dulwich.refs import extract_branch_name

        # Test extracting branch name from full ref
        self.assertEqual(b"master", extract_branch_name(b"refs/heads/master"))
        self.assertEqual(b"develop", extract_branch_name(b"refs/heads/develop"))
        self.assertEqual(
            b"feature/new-ui", extract_branch_name(b"refs/heads/feature/new-ui")
        )

        # Test error on invalid ref
        with self.assertRaises(ValueError) as cm:
            extract_branch_name(b"refs/tags/v1.0")
        self.assertIn("Not a local branch ref", str(cm.exception))

        with self.assertRaises(ValueError):
            extract_branch_name(b"master")

    def test_extract_tag_name(self) -> None:
        """Test extract_tag_name function."""
        from dulwich.refs import extract_tag_name

        # Test extracting tag name from full ref
        self.assertEqual(b"v1.0", extract_tag_name(b"refs/tags/v1.0"))
        self.assertEqual(b"release-2.0", extract_tag_name(b"refs/tags/release-2.0"))

        # Test error on invalid ref
        with self.assertRaises(ValueError) as cm:
            extract_tag_name(b"refs/heads/master")
        self.assertIn("Not a local tag ref", str(cm.exception))

        with self.assertRaises(ValueError):
            extract_tag_name(b"v1.0")


class NamespacedRefsContainerTests(TestCase):
    """Tests for NamespacedRefsContainer."""

    def setUp(self) -> None:
        TestCase.setUp(self)
        # Create an underlying refs container
        self._underlying_refs = DictRefsContainer(dict(_TEST_REFS))
        # Create a namespaced view
        self._refs = NamespacedRefsContainer(self._underlying_refs, b"foo")

    def test_namespace_prefix_simple(self) -> None:
        """Test simple namespace prefix."""
        refs = NamespacedRefsContainer(self._underlying_refs, b"foo")
        self.assertEqual(b"refs/namespaces/foo/", refs._namespace_prefix)

    def test_namespace_prefix_nested(self) -> None:
        """Test nested namespace prefix."""
        refs = NamespacedRefsContainer(self._underlying_refs, b"foo/bar")
        self.assertEqual(
            b"refs/namespaces/foo/refs/namespaces/bar/", refs._namespace_prefix
        )

    def test_allkeys_empty_namespace(self) -> None:
        """Test that newly created namespace has no refs except HEAD."""
        # HEAD is shared across namespaces, so it appears even in empty namespace
        self.assertEqual({b"HEAD"}, self._refs.allkeys())

    def test_setitem_and_getitem(self) -> None:
        """Test setting and getting refs in namespace."""
        sha = b"9" * 40
        self._refs[b"refs/heads/master"] = sha
        self.assertEqual(sha, self._refs[b"refs/heads/master"])

        # Verify it's stored with the namespace prefix in underlying container
        self.assertIn(
            b"refs/namespaces/foo/refs/heads/master", self._underlying_refs.allkeys()
        )
        self.assertEqual(
            sha, self._underlying_refs[b"refs/namespaces/foo/refs/heads/master"]
        )

    def test_head_not_namespaced(self) -> None:
        """Test that HEAD is not namespaced."""
        sha = b"a" * 40
        self._refs[b"HEAD"] = sha
        self.assertEqual(sha, self._refs[b"HEAD"])

        # HEAD should be directly in the underlying container, not namespaced
        self.assertIn(b"HEAD", self._underlying_refs.allkeys())
        self.assertNotIn(b"refs/namespaces/foo/HEAD", self._underlying_refs.allkeys())

    def test_isolation_between_namespaces(self) -> None:
        """Test that different namespaces are isolated."""
        sha1 = b"a" * 40
        sha2 = b"b" * 40

        # Create two different namespaces
        refs_foo = NamespacedRefsContainer(self._underlying_refs, b"foo")
        refs_bar = NamespacedRefsContainer(self._underlying_refs, b"bar")

        # Set ref in foo namespace
        refs_foo[b"refs/heads/master"] = sha1

        # Set ref in bar namespace
        refs_bar[b"refs/heads/master"] = sha2

        # Each namespace should only see its own refs (plus shared HEAD)
        self.assertEqual(sha1, refs_foo[b"refs/heads/master"])
        self.assertEqual(sha2, refs_bar[b"refs/heads/master"])
        self.assertEqual({b"HEAD", b"refs/heads/master"}, refs_foo.allkeys())
        self.assertEqual({b"HEAD", b"refs/heads/master"}, refs_bar.allkeys())

    def test_allkeys_filters_namespace(self) -> None:
        """Test that allkeys only returns refs in the namespace."""
        # Add refs in multiple namespaces
        self._underlying_refs[b"refs/namespaces/foo/refs/heads/master"] = b"a" * 40
        self._underlying_refs[b"refs/namespaces/foo/refs/heads/develop"] = b"b" * 40
        self._underlying_refs[b"refs/namespaces/bar/refs/heads/feature"] = b"c" * 40
        self._underlying_refs[b"refs/heads/global"] = b"d" * 40

        # Only refs in 'foo' namespace should be visible (plus HEAD which is shared)
        foo_refs = NamespacedRefsContainer(self._underlying_refs, b"foo")
        self.assertEqual(
            {b"HEAD", b"refs/heads/master", b"refs/heads/develop"}, foo_refs.allkeys()
        )

    def test_set_symbolic_ref(self) -> None:
        """Test symbolic ref creation in namespace."""
        sha = b"e" * 40
        self._refs[b"refs/heads/develop"] = sha
        self._refs.set_symbolic_ref(b"refs/heads/main", b"refs/heads/develop")

        # Both target and link should be namespaced
        self.assertIn(
            b"refs/namespaces/foo/refs/heads/main", self._underlying_refs.allkeys()
        )
        self.assertEqual(
            b"ref: refs/namespaces/foo/refs/heads/develop",
            self._underlying_refs.read_loose_ref(
                b"refs/namespaces/foo/refs/heads/main"
            ),
        )

    def test_remove_if_equals(self) -> None:
        """Test removing refs from namespace."""
        sha = b"f" * 40
        self._refs[b"refs/heads/temp"] = sha

        # Remove the ref
        self.assertTrue(self._refs.remove_if_equals(b"refs/heads/temp", sha))
        self.assertNotIn(b"refs/heads/temp", self._refs.allkeys())
        self.assertNotIn(
            b"refs/namespaces/foo/refs/heads/temp", self._underlying_refs.allkeys()
        )

    def test_get_packed_refs(self) -> None:
        """Test get_packed_refs returns empty dict for DictRefsContainer."""
        # DictRefsContainer doesn't support packed refs, so just verify
        # the wrapper returns an empty dict
        packed = self._refs.get_packed_refs()
        self.assertEqual({}, packed)

    def test_add_if_new(self) -> None:
        """Test add_if_new in namespace."""
        sha = b"1" * 40
        # Should succeed - ref doesn't exist
        self.assertTrue(self._refs.add_if_new(b"refs/heads/new", sha))
        self.assertEqual(sha, self._refs[b"refs/heads/new"])

        # Should fail - ref already exists
        self.assertFalse(self._refs.add_if_new(b"refs/heads/new", b"2" * 40))
        self.assertEqual(sha, self._refs[b"refs/heads/new"])

    def test_set_if_equals(self) -> None:
        """Test set_if_equals in namespace."""
        sha1 = b"a" * 40
        sha2 = b"b" * 40
        self._refs[b"refs/heads/test"] = sha1

        # Should fail with wrong old value
        self.assertFalse(self._refs.set_if_equals(b"refs/heads/test", b"c" * 40, sha2))
        self.assertEqual(sha1, self._refs[b"refs/heads/test"])

        # Should succeed with correct old value
        self.assertTrue(self._refs.set_if_equals(b"refs/heads/test", sha1, sha2))
        self.assertEqual(sha2, self._refs[b"refs/heads/test"])



# test_bitmap.py -- Tests for bitmap support
# Copyright (C) 2025 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Tests for bitmap support."""

import os
import shutil
import tempfile
import unittest
from io import BytesIO

from dulwich.bitmap import (
    BITMAP_OPT_FULL_DAG,
    BITMAP_OPT_HASH_CACHE,
    BITMAP_OPT_LOOKUP_TABLE,
    BITMAP_SIGNATURE,
    BITMAP_VERSION,
    BitmapEntry,
    EWAHBitmap,
    PackBitmap,
    _encode_ewah_words,
    read_bitmap_file,
    write_bitmap_file,
)
from dulwich.object_store import BitmapReachability, GraphTraversalReachability


class EWAHCompressionTests(unittest.TestCase):
    """Tests for EWAH compression helper functions."""

    def test_encode_empty_words(self):
        """Test encoding empty word list."""
        result = _encode_ewah_words([])
        self.assertEqual([], result)

    def test_encode_single_literal(self):
        """Test encoding single literal word."""
        result = _encode_ewah_words([0x123])
        # Should be: RLW(0 run, 1 literal) + literal
        self.assertEqual(2, len(result))
        # RLW bit layout: [literal_words(31)][running_len(32)][running_bit(1)]
        # running_bit=0, running_len=0, literal_words=1
        expected_rlw = (1 << 33) | (0 << 1) | 0
        self.assertEqual(expected_rlw, result[0])
        self.assertEqual(0x123, result[1])

    def test_encode_zero_run(self):
        """Test encoding run of zeros."""
        result = _encode_ewah_words([0, 0, 0])
        # Should be: RLW(3 zeros, 0 literals)
        self.assertEqual(1, len(result))
        # RLW bit layout: [literal_words(31)][running_len(32)][running_bit(1)]
        # running_bit=0, running_len=3, literal_words=0
        expected_rlw = (0 << 33) | (3 << 1) | 0
        self.assertEqual(expected_rlw, result[0])

    def test_encode_ones_run(self):
        """Test encoding run of all-ones."""
        result = _encode_ewah_words([0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF])
        # Should be: RLW(2 ones, 0 literals)
        self.assertEqual(1, len(result))
        # RLW bit layout: [literal_words(31)][running_len(32)][running_bit(1)]
        # running_bit=1, running_len=2, literal_words=0
        expected_rlw = (0 << 33) | (2 << 1) | 1
        self.assertEqual(expected_rlw, result[0])

    def test_encode_run_followed_by_literals(self):
        """Test encoding run followed by literal words."""
        result = _encode_ewah_words([0, 0, 0x123, 0x456])
        # Should be: RLW(2 zeros, 2 literals) + literals
        self.assertEqual(3, len(result))
        # RLW bit layout: [literal_words(31)][running_len(32)][running_bit(1)]
        # running_bit=0, running_len=2, literal_words=2
        expected_rlw = (2 << 33) | (2 << 1) | 0
        self.assertEqual(expected_rlw, result[0])
        self.assertEqual(0x123, result[1])
        self.assertEqual(0x456, result[2])

    def test_encode_mixed_pattern(self):
        """Test encoding mixed runs and literals."""
        result = _encode_ewah_words(
            [0, 0, 0x123, 0xFFFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF]
        )
        # Should be: RLW(2 zeros, 1 literal) + literal + RLW(2 ones, 0 literals)
        self.assertEqual(3, len(result))


class EWAHCompatibilityTests(unittest.TestCase):
    """Tests for EWAH encode/decode compatibility."""

    def test_encode_decode_sparse_bitmap(self):
        """Test encoding and decoding a sparse bitmap with runs."""
        # Create a bitmap with a pattern that benefits from run-length encoding
        # Bits: 0, 1, 128 (has runs of zeros between set bits)
        bitmap = EWAHBitmap()
        bitmap.add(0)
        bitmap.add(1)
        bitmap.add(128)

        # Encode
        encoded = bitmap.encode()

        # Decode
        bitmap2 = EWAHBitmap(encoded)

        # Verify all bits are preserved
        self.assertEqual(len(bitmap), len(bitmap2))
        self.assertIn(0, bitmap2)
        self.assertIn(1, bitmap2)
        self.assertIn(128, bitmap2)
        self.assertNotIn(2, bitmap2)
        self.assertNotIn(127, bitmap2)

    def test_encode_decode_dense_bitmap(self):
        """Test encoding and decoding a dense bitmap."""
        # Create a bitmap with many consecutive bits set
        bitmap = EWAHBitmap()
        for i in range(64):
            bitmap.add(i)

        # Encode
        encoded = bitmap.encode()

        # Decode
        bitmap2 = EWAHBitmap(encoded)

        # Verify all bits are preserved
        self.assertEqual(64, len(bitmap2))
        for i in range(64):
            self.assertIn(i, bitmap2)
        self.assertNotIn(64, bitmap2)

    def test_encode_decode_runs_of_zeros(self):
        """Test encoding and decoding bitmap with long runs of zeros."""
        # Create bitmap: bits 0, 200, 400 (lots of zeros in between)
        bitmap = EWAHBitmap()
        bitmap.add(0)
        bitmap.add(200)
        bitmap.add(400)

        # Encode
        encoded = bitmap.encode()

        # Decode
        bitmap2 = EWAHBitmap(encoded)

        # Verify
        self.assertEqual(3, len(bitmap2))
        self.assertIn(0, bitmap2)
        self.assertIn(200, bitmap2)
        self.assertIn(400, bitmap2)
        self.assertNotIn(1, bitmap2)
        self.assertNotIn(199, bitmap2)
        self.assertNotIn(201, bitmap2)

    def test_encode_decode_mixed_pattern(self):
        """Test encoding and decoding bitmap with mixed dense/sparse regions."""
        bitmap = EWAHBitmap()
        # Dense region: 0-63
        for i in range(64):
            bitmap.add(i)
        # Sparse region: 200, 300, 400
        bitmap.add(200)
        bitmap.add(300)
        bitmap.add(400)
        # Another dense region: 500-563
        for i in range(500, 564):
            bitmap.add(i)

        # Encode
        encoded = bitmap.encode()

        # Decode
        bitmap2 = EWAHBitmap(encoded)

        # Verify all bits preserved
        self.assertEqual(len(bitmap), len(bitmap2))
        for i in range(64):
            self.assertIn(i, bitmap2)
        self.assertIn(200, bitmap2)
        self.assertIn(300, bitmap2)
        self.assertIn(400, bitmap2)
        for i in range(500, 564):
            self.assertIn(i, bitmap2)
        # Check some bits that shouldn't be set
        self.assertNotIn(64, bitmap2)
        self.assertNotIn(199, bitmap2)
        self.assertNotIn(201, bitmap2)
        self.assertNotIn(499, bitmap2)

    def test_encode_decode_preserves_bitwise_ops(self):
        """Test that encode/decode doesn't break bitwise operations."""
        # Create two bitmaps
        bitmap1 = EWAHBitmap()
        bitmap1.add(0)
        bitmap1.add(5)
        bitmap1.add(100)

        bitmap2 = EWAHBitmap()
        bitmap2.add(5)
        bitmap2.add(10)
        bitmap2.add(100)

        # Encode and decode both
        bitmap1_decoded = EWAHBitmap(bitmap1.encode())
        bitmap2_decoded = EWAHBitmap(bitmap2.encode())

        # Perform operations on decoded bitmaps
        or_result = bitmap1_decoded | bitmap2_decoded
        and_result = bitmap1_decoded & bitmap2_decoded
        xor_result = bitmap1_decoded ^ bitmap2_decoded

        # Verify results
        # OR: {0, 5, 10, 100}
        self.assertEqual(4, len(or_result))
        self.assertIn(0, or_result)
        self.assertIn(5, or_result)
        self.assertIn(10, or_result)
        self.assertIn(100, or_result)

        # AND: {5, 100}
        self.assertEqual(2, len(and_result))
        self.assertIn(5, and_result)
        self.assertIn(100, and_result)

        # XOR: {0, 10}
        self.assertEqual(2, len(xor_result))
        self.assertIn(0, xor_result)
        self.assertIn(10, xor_result)

    def test_round_trip_large_bitmap(self):
        """Test round-trip encoding/decoding of a large bitmap."""
        # Create a large bitmap with various patterns
        bitmap = EWAHBitmap()
        # Add bits at various intervals
        for i in range(0, 10000, 7):
            bitmap.add(i)

        original_bits = set(bitmap.bits)

        # Encode and decode
        encoded = bitmap.encode()
        decoded = EWAHBitmap(encoded)

        # Verify all bits preserved
        self.assertEqual(original_bits, decoded.bits)
        self.assertEqual(len(bitmap), len(decoded))


class EWAHBitmapTests(unittest.TestCase):
    """Tests for EWAH bitmap compression."""

    def test_empty_bitmap(self):
        """Test empty bitmap."""
        bitmap = EWAHBitmap()
        self.assertEqual(0, len(bitmap))
        self.assertEqual(0, bitmap.bit_count)

    def test_add_bit(self):
        """Test adding bits to bitmap."""
        bitmap = EWAHBitmap()
        bitmap.add(0)
        bitmap.add(5)
        bitmap.add(100)

        self.assertEqual(3, len(bitmap))
        self.assertEqual(101, bitmap.bit_count)
        self.assertIn(0, bitmap)
        self.assertIn(5, bitmap)
        self.assertIn(100, bitmap)
        self.assertNotIn(1, bitmap)
        self.assertNotIn(99, bitmap)

    def test_encode_decode(self):
        """Test encoding and decoding bitmaps."""
        bitmap = EWAHBitmap()
        bitmap.add(0)
        bitmap.add(1)
        bitmap.add(64)
        bitmap.add(128)

        # Encode
        data = bitmap.encode()
        self.assertIsInstance(data, bytes)

        # Decode
        bitmap2 = EWAHBitmap(data)
        self.assertEqual(len(bitmap), len(bitmap2))
        self.assertIn(0, bitmap2)
        self.assertIn(1, bitmap2)
        self.assertIn(64, bitmap2)
        self.assertIn(128, bitmap2)

    def test_bitwise_or(self):
        """Test bitwise OR operation."""
        bitmap1 = EWAHBitmap()
        bitmap1.add(0)
        bitmap1.add(5)

        bitmap2 = EWAHBitmap()
        bitmap2.add(5)
        bitmap2.add(10)

        result = bitmap1 | bitmap2
        self.assertEqual(3, len(result))
        self.assertIn(0, result)
        self.assertIn(5, result)
        self.assertIn(10, result)

    def test_bitwise_and(self):
        """Test bitwise AND operation."""
        bitmap1 = EWAHBitmap()
        bitmap1.add(0)
        bitmap1.add(5)

        bitmap2 = EWAHBitmap()
        bitmap2.add(5)
        bitmap2.add(10)

        result = bitmap1 & bitmap2
        self.assertEqual(1, len(result))
        self.assertIn(5, result)
        self.assertNotIn(0, result)
        self.assertNotIn(10, result)

    def test_bitwise_xor(self):
        """Test bitwise XOR operation."""
        bitmap1 = EWAHBitmap()
        bitmap1.add(0)
        bitmap1.add(5)

        bitmap2 = EWAHBitmap()
        bitmap2.add(5)
        bitmap2.add(10)

        result = bitmap1 ^ bitmap2
        self.assertEqual(2, len(result))
        self.assertIn(0, result)
        self.assertIn(10, result)
        self.assertNotIn(5, result)


class BitmapEntryTests(unittest.TestCase):
    """Tests for bitmap entries."""

    def test_create_entry(self):
        """Test creating a bitmap entry."""
        bitmap = EWAHBitmap()
        bitmap.add(0)
        bitmap.add(10)

        entry = BitmapEntry(
            object_pos=100,
            xor_offset=0,
            flags=0,
            bitmap=bitmap,
        )

        self.assertEqual(100, entry.object_pos)
        self.assertEqual(0, entry.xor_offset)
        self.assertEqual(0, entry.flags)
        self.assertEqual(bitmap, entry.bitmap)


class PackBitmapTests(unittest.TestCase):
    """Tests for pack bitmap."""

    def test_create_bitmap(self):
        """Test creating a pack bitmap."""
        bitmap = PackBitmap()
        self.assertEqual(BITMAP_VERSION, bitmap.version)
        self.assertEqual(BITMAP_OPT_FULL_DAG, bitmap.flags)
        self.assertIsNone(bitmap.pack_checksum)

    def test_bitmap_with_entries(self):
        """Test bitmap with entries."""
        bitmap = PackBitmap()
        commit_sha = b"\x00" * 20

        ewah_bitmap = EWAHBitmap()
        ewah_bitmap.add(0)
        ewah_bitmap.add(5)

        entry = BitmapEntry(
            object_pos=0,
            xor_offset=0,
            flags=0,
            bitmap=ewah_bitmap,
        )

        bitmap.entries[commit_sha] = entry

        self.assertTrue(bitmap.has_commit(commit_sha))
        self.assertFalse(bitmap.has_commit(b"\x01" * 20))

    def test_iter_commits(self):
        """Test iterating over commits with bitmaps."""
        bitmap = PackBitmap()
        commit1 = b"\x00" * 20
        commit2 = b"\x01" * 20

        ewah_bitmap = EWAHBitmap()
        entry = BitmapEntry(0, 0, 0, ewah_bitmap)

        bitmap.entries[commit1] = entry
        bitmap.entries[commit2] = entry

        commits = list(bitmap.iter_commits())
        self.assertEqual(2, len(commits))
        self.assertIn(commit1, commits)
        self.assertIn(commit2, commits)


class BitmapFileTests(unittest.TestCase):
    """Tests for bitmap file I/O."""

    def test_write_read_empty_bitmap(self):
        """Test writing and reading an empty bitmap."""
        bitmap = PackBitmap()
        bitmap.pack_checksum = b"\x00" * 20

        # Write to bytes
        f = BytesIO()
        write_bitmap_file(f, bitmap)

        # Read back
        f.seek(0)
        bitmap2 = read_bitmap_file(f)

        self.assertEqual(bitmap.version, bitmap2.version)
        self.assertEqual(bitmap.flags, bitmap2.flags)
        self.assertEqual(bitmap.pack_checksum, bitmap2.pack_checksum)

    def test_write_read_with_type_bitmaps(self):
        """Test writing and reading bitmaps with type information."""
        bitmap = PackBitmap()
        bitmap.pack_checksum = b"\xaa" * 20

        # Add some type bitmap data
        bitmap.commit_bitmap.add(0)
        bitmap.commit_bitmap.add(5)
        bitmap.tree_bitmap.add(1)
        bitmap.blob_bitmap.add(2)
        bitmap.tag_bitmap.add(3)

        # Write to bytes
        f = BytesIO()
        write_bitmap_file(f, bitmap)

        # Read back
        f.seek(0)
        bitmap2 = read_bitmap_file(f)

        self.assertEqual(bitmap.version, bitmap2.version)
        self.assertEqual(bitmap.pack_checksum, bitmap2.pack_checksum)

        # Check type bitmaps
        self.assertIn(0, bitmap2.commit_bitmap)
        self.assertIn(5, bitmap2.commit_bitmap)
        self.assertIn(1, bitmap2.tree_bitmap)
        self.assertIn(2, bitmap2.blob_bitmap)
        self.assertIn(3, bitmap2.tag_bitmap)

    def test_invalid_signature(self):
        """Test reading file with invalid signature."""
        f = BytesIO(b"XXXX\x00\x01\x00\x00")
        with self.assertRaises(ValueError) as cm:
            read_bitmap_file(f)
        self.assertIn("Invalid bitmap signature", str(cm.exception))

    def test_invalid_version(self):
        """Test reading file with invalid version."""
        f = BytesIO(BITMAP_SIGNATURE + b"\x00\x02\x00\x01")
        with self.assertRaises(ValueError) as cm:
            read_bitmap_file(f)
        self.assertIn("Unsupported bitmap version", str(cm.exception))

    def test_incomplete_header(self):
        """Test reading file with incomplete header."""
        f = BytesIO(BITMAP_SIGNATURE + b"\x00")
        with self.assertRaises(ValueError) as cm:
            read_bitmap_file(f)
        self.assertIn("Incomplete bitmap header", str(cm.exception))


class BitmapIntegrationTests(unittest.TestCase):
    """Integration tests for bitmap functionality."""

    def test_round_trip_file(self):
        """Test writing and reading bitmap to/from actual file."""
        with tempfile.TemporaryDirectory() as tmpdir:
            bitmap_path = os.path.join(tmpdir, "test.bitmap")

            # Create bitmap
            bitmap = PackBitmap()
            bitmap.pack_checksum = b"\xff" * 20
            bitmap.commit_bitmap.add(10)
            bitmap.tree_bitmap.add(20)

            # Write to file
            from dulwich.bitmap import write_bitmap

            write_bitmap(bitmap_path, bitmap)

            # Read back
            from dulwich.bitmap import read_bitmap

            bitmap2 = read_bitmap(bitmap_path)

            self.assertEqual(bitmap.version, bitmap2.version)
            self.assertEqual(bitmap.pack_checksum, bitmap2.pack_checksum)
            self.assertIn(10, bitmap2.commit_bitmap)
            self.assertIn(20, bitmap2.tree_bitmap)

    def test_xor_decompression(self):
        """Test XOR decompression of bitmap entries."""
        bitmap = PackBitmap()

        # Create base bitmap
        base_bitmap = EWAHBitmap()
        base_bitmap.add(0)
        base_bitmap.add(1)
        base_bitmap.add(2)

        base_entry = BitmapEntry(
            object_pos=0,
            xor_offset=0,
            flags=0,
            bitmap=base_bitmap,
        )

        # Create XOR'd bitmap
        # If we XOR base (bits 0,1,2) with XOR bitmap (bits 1,3),
        # result should be (bits 0,2,3)
        xor_bitmap = EWAHBitmap()
        xor_bitmap.add(1)
        xor_bitmap.add(3)

        xor_entry = BitmapEntry(
            object_pos=1,
            xor_offset=1,  # Reference the previous entry
            flags=0,
            bitmap=xor_bitmap,
        )

        # Add entries
        base_sha = b"\x00" * 20
        xor_sha = b"\x01" * 20

        bitmap.entries[base_sha] = base_entry
        bitmap.entries_list.append((base_sha, base_entry))

        bitmap.entries[xor_sha] = xor_entry
        bitmap.entries_list.append((xor_sha, xor_entry))

        # Get decompressed bitmap
        result = bitmap.get_bitmap(xor_sha)

        self.assertIsNotNone(result)
        self.assertIn(0, result)
        self.assertNotIn(1, result)  # XOR cancels this
        self.assertIn(2, result)
        self.assertIn(3, result)

    def test_lookup_table_round_trip(self):
        """Test reading and writing lookup tables."""
        with tempfile.TemporaryDirectory() as tmpdir:
            bitmap_path = os.path.join(tmpdir, "test.bitmap")

            # Create bitmap with lookup table
            bitmap = PackBitmap()
            bitmap.flags = BITMAP_OPT_FULL_DAG | BITMAP_OPT_LOOKUP_TABLE
            bitmap.pack_checksum = b"\xaa" * 20

            # Add some bitmap entries (required for lookup table to be written/read)
            for i in range(3):
                ewah = EWAHBitmap()
                ewah.add(i)
                entry = BitmapEntry(
                    object_pos=i,
                    xor_offset=0,
                    flags=0,
                    bitmap=ewah,
                )
                sha = i.to_bytes(20, byteorder="big")
                bitmap.entries[sha] = entry
                bitmap.entries_list.append((sha, entry))

            bitmap.lookup_table = [
                (0, 100, 0),
                (1, 200, 0),
                (2, 300, 1),
            ]

            # Write to file
            from dulwich.bitmap import write_bitmap

            write_bitmap(bitmap_path, bitmap)

            # Read back
            from dulwich.bitmap import read_bitmap

            bitmap2 = read_bitmap(bitmap_path)

            self.assertEqual(bitmap.flags, bitmap2.flags)
            self.assertIsNotNone(bitmap2.lookup_table)
            self.assertEqual(3, len(bitmap2.lookup_table))
            self.assertEqual((0, 100, 0), bitmap2.lookup_table[0])
            self.assertEqual((1, 200, 0), bitmap2.lookup_table[1])
            self.assertEqual((2, 300, 1), bitmap2.lookup_table[2])

    def test_name_hash_cache_round_trip(self):
        """Test reading and writing name-hash cache."""
        with tempfile.TemporaryDirectory() as tmpdir:
            bitmap_path = os.path.join(tmpdir, "test.bitmap")

            # Create bitmap with name-hash cache
            bitmap = PackBitmap()
            bitmap.flags = BITMAP_OPT_FULL_DAG | BITMAP_OPT_HASH_CACHE
            bitmap.pack_checksum = b"\xbb" * 20
            bitmap.name_hash_cache = [0x12345678, 0x9ABCDEF0, 0xFEDCBA98]

            # Write to file
            from dulwich.bitmap import write_bitmap

            write_bitmap(bitmap_path, bitmap)

            # Read back
            from dulwich.bitmap import read_bitmap

            bitmap2 = read_bitmap(bitmap_path)

            self.assertEqual(bitmap.flags, bitmap2.flags)
            self.assertIsNotNone(bitmap2.name_hash_cache)
            self.assertEqual(3, len(bitmap2.name_hash_cache))
            self.assertEqual(0x12345678, bitmap2.name_hash_cache[0])
            self.assertEqual(0x9ABCDEF0, bitmap2.name_hash_cache[1])
            self.assertEqual(0xFEDCBA98, bitmap2.name_hash_cache[2])


class BitmapErrorHandlingTests(unittest.TestCase):
    """Tests for error handling in bitmap reading."""

    def test_truncated_header(self):
        """Test reading bitmap with truncated header."""
        # Only 10 bytes instead of full header
        data = b"BITM\x00\x01\x00\x00\x00\x00"

        with self.assertRaises(ValueError) as ctx:
            read_bitmap_file(BytesIO(data))

        # Should raise ValueError about missing/incomplete header data
        self.assertIsInstance(ctx.exception, ValueError)

    def test_invalid_signature(self):
        """Test reading bitmap with invalid signature."""
        data = b"JUNK\x00\x01\x00\x00" + b"\x00" * 20

        with self.assertRaises(ValueError) as ctx:
            read_bitmap_file(BytesIO(data))

        self.assertIn("signature", str(ctx.exception).lower())

    def test_unsupported_version(self):
        """Test reading bitmap with unsupported version."""
        data = BITMAP_SIGNATURE
        data += b"\x00\x99"  # Version 153 (unsupported)
        data += b"\x00\x00"  # Flags
        data += b"\x00\x00\x00\x00"  # Entry count
        data += b"\x00" * 20  # Pack checksum

        with self.assertRaises(ValueError) as ctx:
            read_bitmap_file(BytesIO(data))

        self.assertIn("version", str(ctx.exception).lower())

    def test_truncated_type_bitmap(self):
        """Test reading bitmap with truncated type bitmap data."""
        # Valid header
        data = BITMAP_SIGNATURE
        data += BITMAP_VERSION.to_bytes(2, "big")
        data += b"\x00\x00"  # Flags
        data += b"\x00\x00\x00\x00"  # Entry count
        data += b"\x00" * 20  # Pack checksum

        # Truncated type bitmap (incomplete EWAH header)
        data += b"\x00\x00\x00\x05"  # bit_count
        data += b"\x00\x00"  # Incomplete word_count

        with self.assertRaises(ValueError) as ctx:
            read_bitmap_file(BytesIO(data))

        self.assertIn("type bitmap", str(ctx.exception).lower())

    def test_truncated_bitmap_entry(self):
        """Test reading bitmap with truncated entry."""
        # Valid header with 1 entry
        data = BITMAP_SIGNATURE
        data += BITMAP_VERSION.to_bytes(2, "big")
        data += b"\x00\x00"  # Flags
        data += b"\x00\x00\x00\x01"  # 1 entry
        data += b"\x00" * 20  # Pack checksum

        # Write empty type bitmaps
        for _ in range(4):
            empty_ewah = EWAHBitmap().encode()
            data += empty_ewah

        # Truncated entry (only object position, missing rest)
        data += b"\x00\x00\x00\x00"  # Object position
        # Missing: XOR offset, flags, bitmap data

        with self.assertRaises(ValueError) as ctx:
            read_bitmap_file(BytesIO(data))

        # Should raise ValueError about missing data
        self.assertIsInstance(ctx.exception, ValueError)

    def test_empty_bitmap_file(self):
        """Test reading completely empty file."""
        with self.assertRaises(ValueError):
            read_bitmap_file(BytesIO(b""))

    def test_bitmap_with_zero_entries(self):
        """Test valid bitmap with zero entries."""
        bitmap = PackBitmap()
        bitmap.pack_checksum = b"\x00" * 20

        f = BytesIO()
        write_bitmap_file(f, bitmap)
        f.seek(0)

        # Should read successfully
        bitmap2 = read_bitmap_file(f)
        self.assertEqual(0, len(bitmap2.entries))
        self.assertIsNotNone(bitmap2.pack_checksum)


class BitmapEdgeCaseTests(unittest.TestCase):
    """Tests for edge cases in bitmap handling."""

    def test_very_large_bitmap(self):
        """Test bitmap with many bits set."""
        bitmap = EWAHBitmap()
        # Add 100,000 bits
        for i in range(100000):
            if i % 3 == 0:  # Every 3rd bit
                bitmap.add(i)

        # Should encode and decode without issues
        encoded = bitmap.encode()
        decoded = EWAHBitmap(encoded)

        self.assertEqual(len(bitmap), len(decoded))
        # Verify a sample of bits
        self.assertIn(0, decoded)
        self.assertIn(99999, decoded)
        self.assertNotIn(1, decoded)
        self.assertNotIn(99998, decoded)

    def test_bitmap_with_large_gaps(self):
        """Test bitmap with large gaps between set bits."""
        bitmap = EWAHBitmap()
        bitmap.add(0)
        bitmap.add(100000)
        bitmap.add(200000)

        encoded = bitmap.encode()
        decoded = EWAHBitmap(encoded)

        self.assertEqual(3, len(decoded))
        self.assertIn(0, decoded)
        self.assertIn(100000, decoded)
        self.assertIn(200000, decoded)

    def test_bitmap_all_bits_in_word(self):
        """Test bitmap with all 64 bits in a word set."""
        bitmap = EWAHBitmap()
        for i in range(64):
            bitmap.add(i)

        encoded = bitmap.encode()
        decoded = EWAHBitmap(encoded)

        self.assertEqual(64, len(decoded))
        for i in range(64):
            self.assertIn(i, decoded)

    def test_multiple_flags_combined(self):
        """Test bitmap with multiple flags set."""
        bitmap = PackBitmap(
            flags=BITMAP_OPT_FULL_DAG | BITMAP_OPT_HASH_CACHE | BITMAP_OPT_LOOKUP_TABLE
        )
        bitmap.pack_checksum = b"\x00" * 20
        bitmap.lookup_table = [(0, 0, 0), (1, 100, 0)]
        bitmap.name_hash_cache = [0x12345678, 0xABCDEF00]

        # Add an entry
        test_bitmap = EWAHBitmap()
        test_bitmap.add(0)
        test_bitmap.add(5)
        entry = BitmapEntry(object_pos=0, xor_offset=0, flags=0, bitmap=test_bitmap)
        bitmap.entries[b"\x00" * 20] = entry

        # Write and read back
        f = BytesIO()
        write_bitmap_file(f, bitmap)
        f.seek(0)

        bitmap2 = read_bitmap_file(f)

        self.assertEqual(bitmap.flags, bitmap2.flags)
        self.assertTrue(bitmap2.flags & BITMAP_OPT_FULL_DAG)
        self.assertTrue(bitmap2.flags & BITMAP_OPT_HASH_CACHE)
        self.assertTrue(bitmap2.flags & BITMAP_OPT_LOOKUP_TABLE)
        self.assertIsNotNone(bitmap2.lookup_table)
        self.assertIsNotNone(bitmap2.name_hash_cache)


class BitmapConfigTests(unittest.TestCase):
    """Tests for bitmap-related configuration settings."""

    def test_pack_write_bitmaps_default(self):
        """Test pack.writeBitmaps defaults to false."""
        from dulwich.config import ConfigFile

        config = ConfigFile()
        self.assertFalse(config.get_boolean((b"pack",), b"writeBitmaps", False))

    def test_pack_write_bitmaps_true(self):
        """Test pack.writeBitmaps = true."""
        from dulwich.config import ConfigFile

        config = ConfigFile()
        config.set((b"pack",), b"writeBitmaps", b"true")
        self.assertTrue(config.get_boolean((b"pack",), b"writeBitmaps", False))

    def test_pack_write_bitmaps_false(self):
        """Test pack.writeBitmaps = false."""
        from dulwich.config import ConfigFile

        config = ConfigFile()
        config.set((b"pack",), b"writeBitmaps", b"false")
        self.assertFalse(config.get_boolean((b"pack",), b"writeBitmaps", False))

    def test_pack_write_bitmap_hash_cache_default(self):
        """Test pack.writeBitmapHashCache defaults to true."""
        from dulwich.config import ConfigFile

        config = ConfigFile()
        self.assertTrue(config.get_boolean((b"pack",), b"writeBitmapHashCache", True))

    def test_pack_write_bitmap_hash_cache_false(self):
        """Test pack.writeBitmapHashCache = false."""
        from dulwich.config import ConfigFile

        config = ConfigFile()
        config.set((b"pack",), b"writeBitmapHashCache", b"false")
        self.assertFalse(config.get_boolean((b"pack",), b"writeBitmapHashCache", True))

    def test_pack_write_bitmap_lookup_table_default(self):
        """Test pack.writeBitmapLookupTable defaults to true."""
        from dulwich.config import ConfigFile

        config = ConfigFile()
        self.assertTrue(config.get_boolean((b"pack",), b"writeBitmapLookupTable", True))

    def test_repack_write_bitmaps(self):
        """Test repack.writeBitmaps configuration."""
        from dulwich.config import ConfigFile

        config = ConfigFile()
        config.set((b"repack",), b"writeBitmaps", b"true")
        self.assertTrue(config.get_boolean((b"repack",), b"writeBitmaps", False))

    def test_pack_use_bitmap_index_default(self):
        """Test pack.useBitmapIndex defaults to true."""
        from dulwich.config import ConfigFile

        config = ConfigFile()
        self.assertTrue(config.get_boolean((b"pack",), b"useBitmapIndex", True))

    def test_pack_use_bitmap_index_false(self):
        """Test pack.useBitmapIndex = false."""
        from dulwich.config import ConfigFile

        config = ConfigFile()
        config.set((b"pack",), b"useBitmapIndex", b"false")
        self.assertFalse(config.get_boolean((b"pack",), b"useBitmapIndex", True))


class ReachabilityProviderTests(unittest.TestCase):
    """Tests for ObjectReachabilityProvider implementations."""

    def setUp(self):
        """Set up test repository with commits."""
        from dulwich.object_store import DiskObjectStore
        from dulwich.objects import Blob, Commit, Tree

        self.test_dir = tempfile.mkdtemp()
        self.store = DiskObjectStore(self.test_dir)

        # Create a simple commit history:
        # commit1 -> commit2 -> commit3
        #         \-> commit4

        # Create blob and tree
        self.blob1 = Blob.from_string(b"test content 1")
        self.store.add_object(self.blob1)

        self.blob2 = Blob.from_string(b"test content 2")
        self.store.add_object(self.blob2)

        self.tree1 = Tree()
        self.tree1[b"file1.txt"] = (0o100644, self.blob1.id)
        self.store.add_object(self.tree1)

        self.tree2 = Tree()
        self.tree2[b"file1.txt"] = (0o100644, self.blob1.id)
        self.tree2[b"file2.txt"] = (0o100644, self.blob2.id)
        self.store.add_object(self.tree2)

        # Create commit1 (root)
        self.commit1 = Commit()
        self.commit1.tree = self.tree1.id
        self.commit1.message = b"First commit"
        self.commit1.author = self.commit1.committer = b"Test <test@example.com>"
        self.commit1.author_time = self.commit1.commit_time = 1234567890
        self.commit1.author_timezone = self.commit1.commit_timezone = 0
        self.store.add_object(self.commit1)

        # Create commit2 (child of commit1)
        self.commit2 = Commit()
        self.commit2.tree = self.tree1.id
        self.commit2.parents = [self.commit1.id]
        self.commit2.message = b"Second commit"
        self.commit2.author = self.commit2.committer = b"Test <test@example.com>"
        self.commit2.author_time = self.commit2.commit_time = 1234567891
        self.commit2.author_timezone = self.commit2.commit_timezone = 0
        self.store.add_object(self.commit2)

        # Create commit3 (child of commit2)
        self.commit3 = Commit()
        self.commit3.tree = self.tree2.id
        self.commit3.parents = [self.commit2.id]
        self.commit3.message = b"Third commit"
        self.commit3.author = self.commit3.committer = b"Test <test@example.com>"
        self.commit3.author_time = self.commit3.commit_time = 1234567892
        self.commit3.author_timezone = self.commit3.commit_timezone = 0
        self.store.add_object(self.commit3)

        # Create commit4 (child of commit1, creates a branch)
        self.commit4 = Commit()
        self.commit4.tree = self.tree2.id
        self.commit4.parents = [self.commit1.id]
        self.commit4.message = b"Fourth commit"
        self.commit4.author = self.commit4.committer = b"Test <test@example.com>"
        self.commit4.author_time = self.commit4.commit_time = 1234567893
        self.commit4.author_timezone = self.commit4.commit_timezone = 0
        self.store.add_object(self.commit4)

    def tearDown(self):
        """Clean up test directory."""
        import shutil

        # Close store to release file handles on Windows
        self.store.close()
        shutil.rmtree(self.test_dir)

    def test_graph_traversal_reachability_single_commit(self):
        """Test GraphTraversalReachability with single commit."""
        from dulwich.object_store import GraphTraversalReachability

        provider = GraphTraversalReachability(self.store)

        # Get reachable commits from commit1
        reachable = provider.get_reachable_commits(
            [self.commit1.id], exclude=None, shallow=None
        )

        # Should only include commit1
        self.assertEqual({self.commit1.id}, reachable)

    def test_graph_traversal_reachability_linear_history(self):
        """Test GraphTraversalReachability with linear history."""
        from dulwich.object_store import GraphTraversalReachability

        provider = GraphTraversalReachability(self.store)

        # Get reachable commits from commit3
        reachable = provider.get_reachable_commits(
            [self.commit3.id], exclude=None, shallow=None
        )

        # Should include commit3, commit2, and commit1
        expected = {self.commit1.id, self.commit2.id, self.commit3.id}
        self.assertEqual(expected, reachable)

    def test_graph_traversal_reachability_with_exclusion(self):
        """Test GraphTraversalReachability with exclusion."""
        from dulwich.object_store import GraphTraversalReachability

        provider = GraphTraversalReachability(self.store)

        # Get commits reachable from commit3 but not from commit1
        reachable = provider.get_reachable_commits(
            [self.commit3.id], exclude=[self.commit1.id], shallow=None
        )

        # Should include commit3 and commit2, but not commit1
        expected = {self.commit2.id, self.commit3.id}
        self.assertEqual(expected, reachable)

    def test_graph_traversal_reachability_branching(self):
        """Test GraphTraversalReachability with branching history."""
        from dulwich.object_store import GraphTraversalReachability

        provider = GraphTraversalReachability(self.store)

        # Get reachable commits from both commit3 and commit4
        reachable = provider.get_reachable_commits(
            [self.commit3.id, self.commit4.id], exclude=None, shallow=None
        )

        # Should include all commits
        expected = {self.commit1.id, self.commit2.id, self.commit3.id, self.commit4.id}
        self.assertEqual(expected, reachable)

    def test_graph_traversal_reachable_objects(self):
        """Test GraphTraversalReachability.get_reachable_objects()."""
        from dulwich.object_store import GraphTraversalReachability

        provider = GraphTraversalReachability(self.store)

        # Get all objects reachable from commit3
        reachable = provider.get_reachable_objects(
            [self.commit3.id], exclude_commits=None
        )

        # Should include commit3, blob1, and blob2 (but not tree objects themselves)
        self.assertIn(self.commit3.id, reachable)
        self.assertIn(self.blob1.id, reachable)
        self.assertIn(self.blob2.id, reachable)
        # Verify at least 3 objects
        self.assertGreaterEqual(len(reachable), 3)

    def test_graph_traversal_reachable_objects_with_exclusion(self):
        """Test GraphTraversalReachability.get_reachable_objects() with exclusion."""
        from dulwich.object_store import GraphTraversalReachability

        provider = GraphTraversalReachability(self.store)

        # Get objects reachable from commit3 but not from commit2
        reachable = provider.get_reachable_objects(
            [self.commit3.id], exclude_commits=[self.commit2.id]
        )

        # commit2 uses tree1 (which has blob1), commit3 uses tree2 (which has blob1 + blob2)
        # So should include commit3 and blob2 (new in commit3)
        # blob1 should be excluded because it's in tree1 (reachable from commit2)
        self.assertIn(self.commit3.id, reachable)
        self.assertIn(self.blob2.id, reachable)

    def test_get_reachability_provider_without_bitmaps(self):
        """Test get_reachability_provider returns GraphTraversalReachability when no bitmaps."""
        from dulwich.object_store import GraphTraversalReachability

        provider = self.store.get_reachability_provider()

        # Should return GraphTraversalReachability when no bitmaps available
        self.assertIsInstance(provider, GraphTraversalReachability)

    def test_get_reachability_provider_prefer_bitmaps_false(self):
        """Test get_reachability_provider with prefer_bitmaps=False."""
        from dulwich.object_store import GraphTraversalReachability

        provider = self.store.get_reachability_provider(prefer_bitmaps=False)

        # Should return GraphTraversalReachability when prefer_bitmaps=False
        self.assertIsInstance(provider, GraphTraversalReachability)

    def test_bitmap_reachability_fallback_without_bitmaps(self):
        """Test BitmapReachability falls back to graph traversal without bitmaps."""
        provider = BitmapReachability(self.store)

        # Without bitmaps, should fall back to graph traversal
        reachable = provider.get_reachable_commits(
            [self.commit3.id], exclude=None, shallow=None
        )

        # Should still work via fallback
        expected = {self.commit1.id, self.commit2.id, self.commit3.id}
        self.assertEqual(expected, reachable)

    def test_bitmap_reachability_fallback_with_shallow(self):
        """Test BitmapReachability falls back for shallow clones."""
        provider = BitmapReachability(self.store)

        # With shallow boundary, should fall back to graph traversal
        reachable = provider.get_reachable_commits(
            [self.commit3.id], exclude=None, shallow={self.commit2.id}
        )

        # Should include commit3 and commit2 (shallow boundary includes boundary commit)
        # but not commit1 (beyond shallow boundary)
        self.assertEqual({self.commit2.id, self.commit3.id}, reachable)

    def test_reachability_provider_protocol(self):
        """Test that both providers implement the same interface."""
        graph_provider = GraphTraversalReachability(self.store)
        bitmap_provider = BitmapReachability(self.store)

        # Both should have the same methods
        for method in [
            "get_reachable_commits",
            "get_reachable_objects",
            "get_tree_objects",
        ]:
            self.assertTrue(hasattr(graph_provider, method))
            self.assertTrue(hasattr(bitmap_provider, method))

    def test_graph_traversal_vs_bitmap_consistency(self):
        """Test that GraphTraversalReachability and BitmapReachability produce same results."""
        graph_provider = GraphTraversalReachability(self.store)
        bitmap_provider = BitmapReachability(self.store)  # Will use fallback

        # Test get_reachable_commits
        graph_commits = graph_provider.get_reachable_commits(
            [self.commit3.id], exclude=[self.commit1.id], shallow=None
        )
        bitmap_commits = bitmap_provider.get_reachable_commits(
            [self.commit3.id], exclude=[self.commit1.id], shallow=None
        )
        self.assertEqual(graph_commits, bitmap_commits)

        # Test get_reachable_objects
        graph_objects = graph_provider.get_reachable_objects(
            [self.commit3.id], exclude_commits=None
        )
        bitmap_objects = bitmap_provider.get_reachable_objects(
            [self.commit3.id], exclude_commits=None
        )
        self.assertEqual(graph_objects, bitmap_objects)


class PackEnsureBitmapTests(unittest.TestCase):
    """Tests for Pack.ensure_bitmap() method."""

    def setUp(self):
        """Set up test repository with a pack."""
        from dulwich.object_store import DiskObjectStore
        from dulwich.objects import Blob, Commit, Tree

        self.temp_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.temp_dir)

        # Create pack directory
        os.makedirs(os.path.join(self.temp_dir, "pack"))

        self.store = DiskObjectStore(self.temp_dir)
        # Close store before cleanup to release file handles on Windows
        self.addCleanup(self.store.close)

        # Create test objects
        self.blob = Blob.from_string(b"test content")
        self.store.add_object(self.blob)

        self.tree = Tree()
        self.tree.add(b"file.txt", 0o100644, self.blob.id)
        self.store.add_object(self.tree)

        self.commit = Commit()
        self.commit.tree = self.tree.id
        self.commit.author = self.commit.committer = b"Test <test@example.com>"
        self.commit.author_time = self.commit.commit_time = 1234567890
        self.commit.author_timezone = self.commit.commit_timezone = 0
        self.commit.message = b"Test commit"
        self.store.add_object(self.commit)

        # Repack to create a pack
        self.store.repack()
        self.pack = self.store.packs[0]

    def test_ensure_bitmap_creates_bitmap(self):
        """Test that ensure_bitmap creates a bitmap file."""
        # Initially no bitmap
        self.assertFalse(os.path.exists(self.pack._bitmap_path))

        # Ensure bitmap with commit_interval=1 to ensure our single commit is selected
        refs = {b"refs/heads/master": self.commit.id}
        bitmap = self.pack.ensure_bitmap(self.store, refs, commit_interval=1)

        # Bitmap should now exist
        self.assertIsNotNone(bitmap)
        self.assertTrue(os.path.exists(self.pack._bitmap_path))
        # Verify it's a PackBitmap instance
        from dulwich.bitmap import PackBitmap

        self.assertIsInstance(bitmap, PackBitmap)

    def test_ensure_bitmap_returns_existing(self):
        """Test that ensure_bitmap returns existing bitmap without regenerating."""
        refs = {b"refs/heads/master": self.commit.id}

        # Create bitmap with commit_interval=1
        self.pack.ensure_bitmap(self.store, refs, commit_interval=1)
        mtime1 = os.path.getmtime(self.pack._bitmap_path)

        # Ensure again - should return existing
        import time

        time.sleep(0.01)  # Ensure time difference
        self.pack.ensure_bitmap(self.store, refs, commit_interval=1)
        mtime2 = os.path.getmtime(self.pack._bitmap_path)

        # File should not have been regenerated
        self.assertEqual(mtime1, mtime2)

    def test_ensure_bitmap_with_custom_interval(self):
        """Test ensure_bitmap with custom commit_interval."""
        refs = {b"refs/heads/master": self.commit.id}
        bitmap = self.pack.ensure_bitmap(self.store, refs, commit_interval=50)
        self.assertIsNotNone(bitmap)


class GeneratePackBitmapsTests(unittest.TestCase):
    """Tests for PackBasedObjectStore.generate_pack_bitmaps()."""

    def setUp(self):
        """Set up test repository."""
        from dulwich.object_store import DiskObjectStore
        from dulwich.objects import Blob, Commit, Tree

        self.temp_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.temp_dir)

        # Create pack directory
        os.makedirs(os.path.join(self.temp_dir, "pack"))

        self.store = DiskObjectStore(self.temp_dir)
        # Close store before cleanup to release file handles on Windows
        self.addCleanup(self.store.close)

        # Create multiple commits
        self.commits = []
        for i in range(3):
            blob = Blob.from_string(f"content {i}".encode())
            self.store.add_object(blob)

            tree = Tree()
            tree.add(f"file{i}.txt".encode(), 0o100644, blob.id)
            self.store.add_object(tree)

            commit = Commit()
            commit.tree = tree.id
            if i > 0:
                commit.parents = [self.commits[-1].id]
            commit.author = commit.committer = b"Test <test@example.com>"
            commit.author_time = commit.commit_time = 1234567890 + i
            commit.author_timezone = commit.commit_timezone = 0
            commit.message = f"Commit {i}".encode()
            self.store.add_object(commit)
            self.commits.append(commit)

        # Repack to create pack
        self.store.repack()

    def test_generate_pack_bitmaps(self):
        """Test generating bitmaps for all packs."""
        refs = {b"refs/heads/master": self.commits[-1].id}

        # Initially no bitmaps
        for pack in self.store.packs:
            self.assertFalse(os.path.exists(pack._bitmap_path))

        # Generate bitmaps
        count = self.store.generate_pack_bitmaps(refs)

        # Should have generated bitmaps
        self.assertEqual(count, len(self.store.packs))
        for pack in self.store.packs:
            self.assertTrue(os.path.exists(pack._bitmap_path))

    def test_generate_pack_bitmaps_multiple_calls(self):
        """Test that calling generate_pack_bitmaps multiple times is safe."""
        refs = {b"refs/heads/master": self.commits[-1].id}

        # Generate once
        self.store.generate_pack_bitmaps(refs)
        mtimes1 = [os.path.getmtime(p._bitmap_path) for p in self.store.packs]

        # Generate again
        import time

        time.sleep(0.01)
        self.store.generate_pack_bitmaps(refs)
        mtimes2 = [os.path.getmtime(p._bitmap_path) for p in self.store.packs]

        # Should not regenerate existing bitmaps
        self.assertEqual(mtimes1, mtimes2)

    def test_generate_pack_bitmaps_with_progress(self):
        """Test generate_pack_bitmaps with progress callback."""
        refs = {b"refs/heads/master": self.commits[-1].id}
        messages = []

        def progress(msg):
            messages.append(msg)

        self.store.generate_pack_bitmaps(refs, progress=progress)

        # Should have received progress messages
        self.assertGreater(len(messages), 0)



# test_am.py -- test am compatibility with CGit
# Copyright (C) 2025 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Tests for git am compatibility with CGit."""

import os
import shutil
import tempfile

from dulwich import porcelain
from dulwich.repo import Repo

from .utils import CompatTestCase, rmtree_ro, run_git_or_fail


class CompatAmTestCase(CompatTestCase):
    def setUp(self) -> None:
        super().setUp()
        self.test_dir = tempfile.mkdtemp()
        self.addCleanup(rmtree_ro, self.test_dir)
        self.repo_path = os.path.join(self.test_dir, "repo")
        self.repo = Repo.init(self.repo_path, mkdir=True)
        self.addCleanup(self.repo.close)
        # Configure git identity for C git operations
        run_git_or_fail(
            ["-C", self.repo_path, "config", "user.email", "test@example.com"]
        )
        run_git_or_fail(["-C", self.repo_path, "config", "user.name", "Test User"])
        run_git_or_fail(["-C", self.repo_path, "config", "core.autocrlf", "false"])

    def test_dulwich_am_cgit_format_patch(self) -> None:
        """Apply a patch generated by C git format-patch using dulwich am."""
        # Create initial commit with C git
        file_path = os.path.join(self.repo_path, "test.txt")
        with open(file_path, "wb") as f:
            f.write(b"line 1\nline 2\nline 3\n")
        run_git_or_fail(["-C", self.repo_path, "add", "test.txt"])
        run_git_or_fail(["-C", self.repo_path, "commit", "-m", "initial"])

        # Make a change and commit with C git
        with open(file_path, "wb") as f:
            f.write(b"line 1\nline two\nline 3\n")
        run_git_or_fail(["-C", self.repo_path, "add", "test.txt"])
        run_git_or_fail(["-C", self.repo_path, "commit", "-m", "Change line 2"])

        # Generate patch with C git format-patch
        patch_dir = os.path.join(self.test_dir, "patches")
        os.makedirs(patch_dir)
        run_git_or_fail(["-C", self.repo_path, "format-patch", "-1", "-o", patch_dir])
        patch_files = sorted(os.listdir(patch_dir))
        self.assertEqual(len(patch_files), 1)
        patch_path = os.path.join(patch_dir, patch_files[0])

        # Make a copy of the repo at the initial commit
        copy_path = os.path.join(self.test_dir, "copy")
        shutil.copytree(self.repo_path, copy_path)
        run_git_or_fail(["-C", copy_path, "reset", "--hard", "HEAD~1"])

        # Apply with dulwich am
        shas = porcelain.am(copy_path, patches=patch_path)
        self.assertEqual(len(shas), 1)

        # Verify file content matches
        with open(os.path.join(copy_path, "test.txt"), "rb") as f:
            self.assertEqual(f.read(), b"line 1\nline two\nline 3\n")

        # Verify commit message
        copy_repo = Repo(copy_path)
        self.addCleanup(copy_repo.close)
        commit = copy_repo[shas[0]]
        self.assertIn(b"Change line 2", commit.message)

    def test_cgit_am_dulwich_format_patch(self) -> None:
        """Apply a patch generated by dulwich format-patch using C git am."""
        # Create initial commit
        file_path = os.path.join(self.repo_path, "test.txt")
        with open(file_path, "wb") as f:
            f.write(b"line 1\nline 2\nline 3\n")
        porcelain.add(self.repo_path, paths=["test.txt"])
        porcelain.commit(
            self.repo_path,
            message=b"initial",
            author=b"Test User <test@example.com>",
            committer=b"Test User <test@example.com>",
        )

        # Make a change and commit
        with open(file_path, "wb") as f:
            f.write(b"line 1\nline two\nline 3\n")
        porcelain.add(self.repo_path, paths=["test.txt"])
        porcelain.commit(
            self.repo_path,
            message=b"Change line 2",
            author=b"Test User <test@example.com>",
            committer=b"Test User <test@example.com>",
        )

        # Generate patch with dulwich format-patch
        patch_dir = os.path.join(self.test_dir, "patches")
        os.makedirs(patch_dir)
        patch_files = porcelain.format_patch(self.repo_path, n=1, outdir=patch_dir)
        self.assertEqual(len(patch_files), 1)

        # Make a copy at the initial commit
        copy_path = os.path.join(self.test_dir, "copy")
        shutil.copytree(self.repo_path, copy_path)
        run_git_or_fail(["-C", copy_path, "reset", "--hard", "HEAD~1"])

        # Apply with C git am
        run_git_or_fail(["-C", copy_path, "am", patch_files[0]])

        # Verify file content
        with open(os.path.join(copy_path, "test.txt"), "rb") as f:
            self.assertEqual(f.read(), b"line 1\nline two\nline 3\n")

        # Verify commit message
        output = run_git_or_fail(["-C", copy_path, "log", "-1", "--format=%s"])
        self.assertEqual(output.strip(), b"Change line 2")

    def test_dulwich_am_multiple_cgit_patches(self) -> None:
        """Apply multiple patches from C git format-patch using dulwich am."""
        # Create initial commit
        file_path = os.path.join(self.repo_path, "test.txt")
        with open(file_path, "wb") as f:
            f.write(b"line 1\nline 2\nline 3\n")
        run_git_or_fail(["-C", self.repo_path, "add", "test.txt"])
        run_git_or_fail(["-C", self.repo_path, "commit", "-m", "initial"])

        # Two changes
        with open(file_path, "wb") as f:
            f.write(b"line 1\nline two\nline 3\n")
        run_git_or_fail(["-C", self.repo_path, "add", "test.txt"])
        run_git_or_fail(["-C", self.repo_path, "commit", "-m", "Change line 2"])

        with open(file_path, "wb") as f:
            f.write(b"line 1\nline two\nline three\n")
        run_git_or_fail(["-C", self.repo_path, "add", "test.txt"])
        run_git_or_fail(["-C", self.repo_path, "commit", "-m", "Change line 3"])

        # Generate patches with C git
        patch_dir = os.path.join(self.test_dir, "patches")
        os.makedirs(patch_dir)
        run_git_or_fail(["-C", self.repo_path, "format-patch", "-2", "-o", patch_dir])
        patch_files = sorted(os.path.join(patch_dir, f) for f in os.listdir(patch_dir))
        self.assertEqual(len(patch_files), 2)

        # Copy repo at initial commit
        copy_path = os.path.join(self.test_dir, "copy")
        shutil.copytree(self.repo_path, copy_path)
        run_git_or_fail(["-C", copy_path, "reset", "--hard", "HEAD~2"])

        # Apply with dulwich am
        shas = porcelain.am(copy_path, patches=patch_files)
        self.assertEqual(len(shas), 2)

        # Verify final content
        with open(os.path.join(copy_path, "test.txt"), "rb") as f:
            self.assertEqual(f.read(), b"line 1\nline two\nline three\n")

    def test_dulwich_am_preserves_author(self) -> None:
        """Verify dulwich am preserves author info from C git patches."""
        # Create commit with specific author
        file_path = os.path.join(self.repo_path, "test.txt")
        with open(file_path, "wb") as f:
            f.write(b"hello\n")
        run_git_or_fail(["-C", self.repo_path, "add", "test.txt"])
        run_git_or_fail(["-C", self.repo_path, "commit", "-m", "initial"])

        with open(file_path, "wb") as f:
            f.write(b"hello world\n")
        run_git_or_fail(["-C", self.repo_path, "add", "test.txt"])
        run_git_or_fail(
            [
                "-C",
                self.repo_path,
                "commit",
                "--author=Patch Author <author@example.org>",
                "-m",
                "Add world",
            ]
        )

        # Generate patch
        patch_dir = os.path.join(self.test_dir, "patches")
        os.makedirs(patch_dir)
        run_git_or_fail(["-C", self.repo_path, "format-patch", "-1", "-o", patch_dir])
        patch_files = sorted(os.listdir(patch_dir))
        patch_path = os.path.join(patch_dir, patch_files[0])

        # Copy at initial commit
        copy_path = os.path.join(self.test_dir, "copy")
        shutil.copytree(self.repo_path, copy_path)
        run_git_or_fail(["-C", copy_path, "reset", "--hard", "HEAD~1"])

        # Apply with dulwich am
        shas = porcelain.am(copy_path, patches=patch_path)
        self.assertEqual(len(shas), 1)

        # Verify author is preserved
        copy_repo = Repo(copy_path)
        self.addCleanup(copy_repo.close)
        commit = copy_repo[shas[0]]
        self.assertEqual(commit.author, b"Patch Author <author@example.org>")
